{ This library is free software; you can redistribute it and/or modify it
  under the terms of the GNU Library General Public License as published by
  the Free Software Foundation; either version 2 of the License, or (at your
  option) any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Library General Public License
  for more details.

  You should have received a copy of the GNU Library General Public License
  along with this library; if not, write to the Free Software Foundation,
  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

  ---------------------------------------------------------------------------

    Author : Luc DAVID Email: luckylazarus@free.fr
    Copyright (C) 2006 - 2008

    Major changes are indicated in the \Documentation\Changes.pdf file
    Last update 26.02.2008

  --------------------------------------------------------------------------- }


{ TSqlitePassDatabase }

Constructor TSqlitePassDatabase.Create(AOwner : TComponent);
begin
  Inherited Create(AOwner);
  {$IFDEF DEBUG}
  Debugger := TSqlitePassDebugger.Create;
  {$ENDIF}

  FState := dbsInactive;
  { Events }
  FBeforeConnect      := Nil;
  FAfterConnect       := Nil;
  FBeforeDisconnect   := Nil;
  FAfterDisconnect    := Nil;
  FDataTypeConversion := Nil;
  { Properties }
  FActive := False;
  FReadOnly := False;
  FShowSysObjects := False;
  FSmartOpen := True;
  FDatabaseType := dbtUnknown;
  FQueryTimeout := 0;
  { First, creates the appropriate DbEngine }
  FEngine := TSqlitePassEngine.Create;
  FVersionInfo := TSqlitePassDatabaseVersionInfo.Create(Self);
  FDataTypeOptions := TSqlitePassDatabaseDataTypeOptions.Create(Self);
  FOptions := TSqlitePassDatabaseOptions.Create(Self);
  FTableDefs := TSqlitePassTableDefs.Create(Self);
  FQueryDefs := TSqlitePassQueryDefs.Create(Self);
  FIndexDefs := TSqlitePassDatabaseIndexDefs.Create(Self);
  FViews := TSqlitePassViews.Create(Self);
  FTriggers := TSqlitePassTriggers.Create(Self);
  FDatasets := TSqlitePassDatasets.create(Self);
  FDatabases := TSqlitePassDatabasesAttached.Create(Self);
end;

Destructor  TSqlitePassDatabase.Destroy;
begin
   Connected := False;
   If Assigned(FTranslator)
      then FTranslator.Free;
   FDatabases.Free;
   FDatasets.Free;
   FTriggers.Free;
   FViews.Free;
   FIndexDefs.Free;
   FQueryDefs.Free;
   FTableDefs.Free;
   FOptions.Free;
   FDataTypeOptions.Free;
   FVersionInfo.Free;
   FEngine.Free;
  {$IFDEF DEBUG}
  Debugger.Free;
  {$ENDIF}
 Inherited Destroy;
end;

procedure TSqlitePassDatabase.SetFDatabaseType(const AValue: TSqlitePassDatabaseType);
var
PreviousDatabaseType: TSqlitePassDatabaseType;
begin
FDatabaseType := AValue;

{ Creates the appropriate FieldTranslator }
If Assigned(FTranslator)
   then FTranslator.Free;

Case FDatabaseType of
  dbtSqlitePass  : FTranslator := TSqlitePassTranslator.Create(Self);
  dbtKexi        : FTranslator := TSqlitePassTranslator_Kexi.Create(Self);
  dbtSqliteAdmin : FTranslator := TSqlitePassTranslator.Create(Self);
  dbtSqlite4Fpc  : FTranslator := TSqlitePassTranslator.Create(Self);
  dbtSqliteExpert: FTranslator := TSqlitePassTranslator_SqliteExpert.Create(Self);
  else             FTranslator := TSqlitePassTranslator.Create(Self);
  end;

if Not (csLoading in ComponentState)
   then RefreshDefinitions;
end;

Procedure TSqlitePassDatabase.SetFActive(Value : Boolean);
begin
if (FActive <> Value) and not (csloading in ComponentState) then
   begin
   If Value
      then Open
      else Close;
   end;
end;

function TSqlitePassDatabase.CheckCanOpen(ShowError: Boolean = True): Boolean;
var
DatabasePath, Msg: String;

begin
Result := True;
Try
DatabasePath := FDatabase;
  if (DatabasePath='') then
      begin
      Msg := Msg1005;
      Abort;
      end;

  { Gets the database fullpath if FSmartOpen is True }
  if FSmartOpen and (not FileExists(DatabasePath))
    then if csDesigning in ComponentState
            then DatabasePath := GetWorkDir+ExtractFileName(FDatabase)
            else DatabasePath := GetExeDir+ExtractFileName(FDatabase);

  if not FileExists(DatabasePath) then
     begin
      Msg := Msg1038+CRLF+DatabasePath;
      Abort;
      end;

  FEngine.QueryTimeout := FQueryTimeOut;
  FCollatingOrder := GetFCollatingOrder;

  { Tries to open the database and stores the handle to the database }
  Engine.OpenDatabase(DatabasePath, FSqliteLibrary);
  FState := dbsInternalOpen;
Except
  Result := False;
  if ShowError then MessageDlg(Msg, mtWarning, [mbOk], 0);
end;
end;

Procedure TSqlitePassDatabase.Open;
begin
if FActive then Exit;
if Assigned(FBeforeConnect) then FBeforeConnect(Self);

If CheckCanOpen then
   begin
   FActive:=True;
   RefreshDefinitions;
   Idle;
   if Assigned(FAfterConnect) then FAfterConnect(Self);
   FState := dbsActive;
   end;
end;

Procedure TSqlitePassDatabase.Close;
begin
if FActive then
   begin
     if Assigned(FBeforeDisconnect) then FBeforeDisconnect(Self);
     { We save DatatypeOptions to Database if required }
     if (FDatatypeOptions.SaveOptions <> [])
        then FDatatypeOptions.SaveToDatabase;
     { We detach previously attached databases first }
     Databases.DetachAll;
     FDatasets.Close;
     FEngine.CloseDatabase;
     FActive:=False;
     FState := dbsInactive;
     if Assigned(FAfterDisconnect) then FAfterDisconnect(Self);
   end;
end;

{ Gets information about the database tables, queries...
  including attached databases info }
Procedure TSqlitePassDatabase.RefreshDefinitions;
begin
 if Not CheckCanOpen(False) then Exit;

 try
 TableDefs.Refresh;
 QueryDefs.Refresh;
 IndexDefs.Refresh;
 Views.Refresh;
 Triggers.Refresh;
 finally
 If Not Connected then
    begin
    FEngine.CloseDatabase;
    FState := dbsInactive;
    end
    else FState := dbsActive;
 end;
end;

Procedure TSqlitePassDatabase.ClearDefinitions;
begin
  TableDefs.Clear;
  QueryDefs.Clear;
  IndexDefs.Clear;
  Views.Clear;
  Triggers.Clear;
end;

function TSqlitePassDatabase.GetPragma(PragmaName: String): integer;
var
Recordset : TSqlitePassRecordset;
begin
if Not FActive
   then begin
        Result := 0;
        exit;
        end;
Recordset := TSqlitePassRecordset.Create(nil);
With Recordset do
     begin
     Open(Engine.Connection, 'PRAGMA ' + PragmaName + ';');
     Result := StrToInt(Items[0].FieldsValues[0]);
     Free;
     end;
end;

procedure TSqlitePassDatabase.SetPragma(PragmaName: string; value: Boolean);
var
 s:string;
begin
if Assigned(FEngine) and FActive
   then begin
        if Value
           then s:='ON'
           else s:='OFF';
        Engine._ExecSql(Format('PRAGMA %s=%s',[PragmaName, s]));
        end;
end;

procedure TSqlitePassDatabase.SetPragma(PragmaName: string; value: integer);
begin
if Assigned(Engine) and FActive
   then Engine._ExecSql(Format('PRAGMA %s=%d',[PragmaName, Value]));
end;

procedure TSqlitePassDatabase.SetPragma(PragmaName, Value: string);
begin
if Assigned(Engine) and FActive
   then Engine._ExecSql(Format('PRAGMA %s=%s',[PragmaName, Value]));
end;

Procedure TSqlitePassDatabase.Idle;
begin
// TODO : TSqlitePassDatabase.Idle - To be implemented
end;

Procedure TSqlitePassDatabase.SetFDatabaseVersion(Value : String);
begin
  // This property is read only
end;

procedure TSqlitePassDatabase.SetFUser(const AValue: Integer);
begin
  if FUser=AValue then exit;
  FUser:=AValue;
end;

Procedure TSqlitePassDatabase.SetFVersionInfo(Value : String);
begin
  // This property is read only
end;

Function TSqlitePassDatabase.GetFCollatingOrder:String;
begin
Result := '';
// TODO : TSqlitePassDatabase.GetFCollatingOrder - To be implemented
end;

Procedure TSqlitePassDatabase.SetFShowSysObjects(Value : Boolean);
begin
 FShowSysObjects := Value;
end;

Procedure TSqlitePassDatabase.SetFReadOnly(Value : Boolean);
begin
 FReadOnly := Value;
end;

function TSqlitePassDatabase.IsSystemTable(TableName: String): Boolean;
begin
Result := Translator.IsSystemTable(TableName);
end;

function TSqlitePassDatabase.Compact: Integer;
var
f: TFileStream;
OriginalSize: Integer;
begin
f := TFileStream.Create(FDatabase, fmOpenRead or fmShareDenyNone);
OriginalSize := f.Size;
Engine.exec('Vacuum');
Result := OriginalSize - f.Size;
end;

function TSqlitePassDatabase.AttachDatabase(FAttachedDatabase, DatabaseAlias: String): Boolean;
var
AttachedDb: TSqlitePassDatabaseAttached;
ExtPos: Integer;
FileExt: String;
CanAttach: Boolean;
begin
Result := False;
if DatabaseAlias = ''
   then begin
        DatabaseAlias := StringReplace(ExtractFileName(FAttachedDatabase), ' ', '', [rfReplaceAll]);
        ExtPos := Pos('.', DatabaseAlias);
        FileExt := System.Copy(DatabaseAlias, ExtPos, MaxInt);
        DatabaseAlias := System.Copy(DatabaseAlias, 0, Pred(ExtPos));
        end;

CanAttach := (FDatabaseType = dbtKexi) and (FileExt = '.kexi')
          or (FDatabaseType = dbtSqliteAdmin) and (FileExt = '.s3db')
          or (FDatabaseType = dbtSqliteExpert) and (FileExt = '.db3')
          or (FDatabaseType = dbtSqlite4Fpc) and (FileExt = '.db');

If Not CanAttach then
   begin
   Raise Exception.Create(Msg1041 + '('+FileExt+')');
   Exit;
   end;

Try
Engine.exec('ATTACH DATABASE "' + FAttachedDatabase +'" AS "'+ DatabaseAlias +'";');
AttachedDb := TSqlitePassDatabaseAttached.Create(Databases);
AttachedDb.Database := FAttachedDatabase;
AttachedDb.FAliasName := DatabaseAlias;
FDatabases.Add(AttachedDb);
Result := True;
Finally
RefreshDefinitions;
end;
end;

function TSqlitePassDatabase.DetachDatabase(DatabaseAlias: String): Boolean;
var
AttachedDb: TSqlitePassDatabaseAttached;
begin
Result := False;
Try
AttachedDb := FDatabases.DatabaseByName(DatabaseAlias);
if Assigned(AttachedDb) then
   begin
   Engine.exec('DETACH DATABASE "' + DatabaseAlias +'";');
   FDatabases.Remove(AttachedDb);
   AttachedDb.Free;
   Result := True;
   end;
Finally
RefreshDefinitions;
end;
end;

procedure TSqlitePassDatabase.SetFDatabase(Value: String);
begin
  if FActive then
     begin
       Raise Exception.Create(Msg1010);
       Exit;
     end;
  FDatabase:=Value;

  if FSmartOpen then
     begin
          if (Pos('.kexi', FDatabase)) > 0 then SetFDatabaseType(dbtKexi)
     else if (Pos('.s3db', FDatabase)) > 0 then SetFDatabaseType(dbtSqliteAdmin)
     else if (Pos('.db3', FDatabase))  > 0 then SetFDatabaseType(dbtSqliteExpert)
     else if (Pos('.db', FDatabase))   > 0 then SetFDatabaseType(dbtSqlite4Fpc)
     else SetFDatabaseType(dbtUnknown);
     end;
end;

procedure TSqlitePassDatabase.SetFSQLiteLibrary(Value: String);
begin
if (FActive) then
   begin
     Raise Exception.Create(Msg1015);
     Exit;
   end;
FSQLiteLibrary:=Value;
end;

function TSqlitePassDatabase.GetFEngine: TSqlitePassEngine;
begin
 if Assigned(FEngine)
    then Result := FEngine
    else Result := nil;
end;

Procedure TSqlitePassDatabase.CreateDatabase(FullName: String);
begin
if Not FileExists(FullName)
   then With TSqlitePassEngine.Create do
        begin
        OpenDatabase(FullName, FSqliteLibrary);
        CloseDatabase;
        Free;
        end
   else DatabaseErrorFmt(Msg3012, [FullName], Self);
end;


function TSqlitePassDatabase.GetFActive: Boolean;
begin
 Result := FActive;
end;

Procedure TSqlitePassDatabase.Loaded;
begin
try
  inherited Loaded;
  RefreshDefinitions;
except
  Application.HandleException(Self)
end;
end;


{ TSqlitePassDatabaseVersionInfo }

function TSqlitePassDatabaseVersionInfo.GetFSqliteEngine: String;
begin
if Assigned(FDatabase.Engine) and FDatabase.FActive
   then Result := SqliteDbv3_SqliteLibVersion
   else Result := '';
end;

function TSqlitePassDatabaseVersionInfo.GetFUserTag: Integer;
begin
 Result := FDatabase.GetPragma('user_version');
end;

function TSqlitePassDatabaseVersionInfo.GetFDbSchema: Integer;
begin
 Result := FDatabase.GetPragma('schema_version');
end;

procedure TSqlitePassDatabaseVersionInfo.SetFDbVersion(const AValue: String);
begin
// Read only - Used to display infos in object inspector
end;

procedure TSqlitePassDatabaseVersionInfo.SetFPackage(const AValue: String);
begin
// Read only - Used to display infos in object inspector
end;

procedure TSqlitePassDatabaseVersionInfo.SetFDbSchema(const AValue: Integer);
begin
// Read only - Used to display infos in object inspector
end;

procedure TSqlitePassDatabaseVersionInfo.SetFSqliteEngine(const AValue: String);
begin
// Read only - Used to display infos in object inspector
end;

procedure TSqlitePassDatabaseVersionInfo.SetFUserTag(const AValue: Integer);
begin
  FDatabase.SetPragma('user_version', AValue);
  FUserTag := AValue;
end;

constructor TSqlitePassDatabaseVersionInfo.Create(Database: TSqlitePassDatabase);
begin
inherited Create();
FDatabase  := Database;
FComponent := SqlitePassDatabaseVersion;
FPackage   := SqlitePassPackageVersion;
FSchema  := 0;
FDbVersion := '';
GetFSqliteEngine;
end;

destructor TSqlitePassDatabaseVersionInfo.Destroy;
begin
FDatabase := nil;
inherited Destroy;
end;

procedure TSqlitePassDatabaseVersionInfo.SetFComponent(
  const Value: String);
begin
// Read only - Used to display infos in object inspector
end;

{ TSqlitePassDatabaseOptions }

constructor TSqlitePassDatabaseOptions.Create(Database: TSqlitePassDatabase);
begin
inherited Create;
FDatabase := Database;
QuoteStyle := qsDoubleQuote;
CaseSensitiveLike := True;
end;

Destructor TSqlitePassDatabaseOptions.Destroy;
begin
FDatabase := nil;
inherited Destroy;
end;

function TSqlitePassDatabaseOptions.GetFAutoVacuum: Boolean;
begin
 Result := Boolean(FDatabase.GetPragma('auto_vacuum'))
end;

function TSqlitePassDatabaseOptions.GetFCacheSize: Integer;
begin
 Result := FDatabase.GetPragma('cache_size')
end;

function TSqlitePassDatabaseOptions.GetFCountChanges: Boolean;
begin
 Result := Boolean(FDatabase.GetPragma('count_changes'));
end;

function TSqlitePassDatabaseOptions.GetFDefaultCacheSize: Integer;
begin
  Result := FDatabase.GetPragma('default_cache_size');
end;

function TSqlitePassDatabaseOptions.GetFEmptyResultCallBack: Boolean;
begin
 Result := Boolean(FDatabase.GetPragma('empty_result_callbacks'));
end;

function TSqlitePassDatabaseOptions.GetFEncoding: TSqlitePassEncoding;
begin
// TODO : TSqlitePassDatabaseOptions.GetFEncoding - To be implemented
 //Result := FDatabase.GetPragma('encoding');
 Result := UTF_8;
end;

function TSqlitePassDatabaseOptions.GetFFullColumnNames: Boolean;
begin
  Result := Boolean(FDatabase.GetPragma('full_column_names'));
end;

function TSqlitePassDatabaseOptions.GetFPageSize: TSqlitePassPageSize;
begin
  Result := FDatabase.GetPragma('page_size');
end;

function TSqlitePassDatabaseOptions.GetFSynchronous: TSqlitePassSynchronous;
begin
 Result := TSqlitePassSynchronous(FDatabase.GetPragma('synchronous'));
end;

function TSqlitePassDatabaseOptions.GetFTemporaryStorage: TSqlitePassTempStore;
begin
 Result := TSqlitePassTempStore(FDatabase.GetPragma('temp_store'));
end;

procedure TSqlitePassDatabaseOptions.SetFAutoVacuum(const AValue: Boolean);
begin
 FDatabase.SetPragma('auto_vacuum',Avalue);
 FAutoVacuum := AValue;
end;

procedure TSqlitePassDatabaseOptions.SetFCacheSize(const AValue: Integer);
begin
 FDatabase.SetPragma('cache_size',Avalue);
 FCacheSize := AValue;
end;

procedure TSqlitePassDatabaseOptions.SetFCaseSensitiveLike(const AValue: Boolean);
begin
 FDatabase.SetPragma('case_sensitive_like',Avalue);
 FCaseSensitiveLike := AValue;
end;

procedure TSqlitePassDatabaseOptions.SetFCountChanges(const AValue: Boolean);
begin
 FDatabase.SetPragma('count_changes',Avalue);
 FCountChanges := AValue;
end;

procedure TSqlitePassDatabaseOptions.SetFDefaultCacheSize(const AValue: Integer);
begin
 FDatabase.SetPragma('default_cache_size',Avalue);
 FDefaultCacheSize := AValue;
end;

procedure TSqlitePassDatabaseOptions.SetFEmptyResultCallBack(const AValue: Boolean);
begin
 FDatabase.SetPragma('empty_result_callbacks', AValue);
 FEmptyResultCallBack := aValue;
end;

procedure TSqlitePassDatabaseOptions.SetFEncoding(const AValue: TSqlitePassEncoding);
var
s: string;
begin
Case AValue of
UTF_8: s:= 'UTF-8';
UTF_16: s:= 'UTF-16';
UTF_16le: s:= 'UTF-16le';
UTF_16be: s:= 'UTF-16be';
end;
 FDatabase.SetPragma('encoding', s);
 GetFEncoding;
end;

procedure TSqlitePassDatabaseOptions.SetFFullColumnNames(const AValue: Boolean);
begin
 FDatabase.SetPragma('full_column_names', AValue);
 FDatabase.SetPragma('short_column_names', not AValue);
 FFullColumnNames := aValue;
 FShortColumnNames := Not aValue;
end;

procedure TSqlitePassDatabaseOptions.SetFPageSize(AValue: TSqlitePassPageSize);
begin
Case AValue of
  1..512    : AValue := 512;
  513..1024 : AValue := 1024;
  1025..2048: AValue := 2048;
  2049..4096: AValue := 4096;
  4097..8192: AValue := 8192;
  else AValue := 2048;
  end;
 FDatabase.SetPragma('page_size', AValue);
 FPageSize := aValue;
end;

procedure TSqlitePassDatabaseOptions.SetFSynchronous(const AValue: TSqlitePassSynchronous);
begin
 FDatabase.SetPragma('synchronous', Integer(AValue));
 FSynchronous := aValue;
end;

procedure TSqlitePassDatabaseOptions.SetFTemporaryStorage(const AValue: TSqlitePassTempStore);
begin
 FDatabase.SetPragma('temp_store', Integer(AValue));
 FTemporaryStorage := aValue;
end;

procedure TSqlitePassDatabaseOptions.SetFQuoteStyle(const AValue: TSqlitePassQuoteStyle);
begin
 Case AValue of
      qsSingleQuote: begin FQuoteOpen := ''''; FQuoteClose := ''''; end;
      qsDoubleQuote: begin FQuoteOpen := '"'; FQuoteClose := '"'; end;
      qsBracket    : begin FQuoteOpen := '['; FQuoteClose := ']'; end;
      else           begin FQuoteOpen := #0; FQuoteClose := #0; end;
      end;
FQuoteStyle := AValue;
end;


{ TSqlitePassFieldTypesTranslationRules }


constructor TSqlitePassFieldTypesTranslationRules.Create(
  DataTypeOptions: TSqlitePassDatabaseDataTypeOptions);
begin
inherited Create;
FDataTypeOptions := DataTypeOptions;
end;

Destructor TSqlitePassFieldTypesTranslationRules.Destroy;
begin
ClearAndFreeItems;
FDataTypeOptions := nil;
inherited Destroy;
end;

function TSqlitePassFieldTypesTranslationRules.GetItem(Index: Integer): TSqlitePassFieldTypeTranslationRule;
begin
  Result:=TSqlitePassFieldTypeTranslationRule(Inherited Items[Index]);
end;

procedure TSqlitePassFieldTypesTranslationRules.SetItem(Index: Integer;
  FieldTypeTranslationRule: TSqlitePassFieldTypeTranslationRule);
begin
   Put(Index,FieldTypeTranslationRule);
end;

function TSqlitePassFieldTypesTranslationRules.Add(FieldTypeTranslator: TSqlitePassFieldTypeTranslationRule
  ): Integer;
begin
   Result:=Inherited Add(FieldTypeTranslator);
end;

procedure TSqlitePassFieldTypesTranslationRules.ClearAndFreeItems;
var
i: Integer;
begin
{ Clear the internal items }
for i := 0 to Pred(Count)
    do Items[i].Free;
{ Clear the list pointers }
inherited Clear;
end;

Function TSqlitePassFieldTypesTranslationRules.ConvertNativeFieldTypeToPascalFieldType(FieldTypeName: String): TFieldType;
var
i: integer;
begin
Result := ftUnknown;
FieldTypeName := LowerCase(FieldTypeName);
For i:=0 to Pred(Count) do
  begin
  Case Items[i].MappingMode of
       mmExact        : if (Items[i].DatatypeName = FieldTypeName)
                           then Result := Items[i].FieldType;

       mmExactNoCase  : if AnsiCompareText(Items[i].DatatypeName,FieldTypeName) = 0
                           then Result := Items[i].FieldType;

       mmPartial      : if Pos(Items[i].DatatypeName, FieldTypeName) <> 0
                           then Result := Items[i].FieldType;

       mmPartialNoCase: if Pos(Uppercase(Items[i].DatatypeName), Uppercase(FieldTypeName)) <> 0
                           then Result := Items[i].FieldType;

       mmAll          : if (Items[i].DatatypeName = FieldTypeName)
                           then Result := Items[i].FieldType
                           else if Pos(Items[i].DatatypeName, FieldTypeName) <> 0
                                   then Result := Items[i].FieldType;
       end;
  if Result <> ftUnknown
     then Break;
  end;

{ Set a default fieldtype if fieltype is ftUnknown }
If Result = ftUnknown
   then Result := FDataTypeOptions.DefaultFieldType;
end;

Procedure TSqlitePassFieldTypesTranslationRules.SetDefaultTranslationRules(DatabaseType: TSqlitePassDatabaseType);
Type
TSqlitePassPresetFieldType = record
   Name:string;
   Fieldtype:TFieldType;
   end;
   
const
{ Default Fields Type when DatabaseType is dtbUnknown }
DefaultFieldsTypes:array[0..21] of TSqlitePassPresetFieldType =
           ((Name:'widestring'; FieldType:ftString), // must be before string
            (Name:'string'; FieldType:ftString),
            (Name:'smallint'; FieldType:ftSmallInt),     // must be before integer
            (Name:'largeint'; FieldType:ftLargeint),     // must be before integer
            (Name:'integer'; Fieldtype:ftInteger),
            (Name:'word'; FieldType:ftWord),
            (Name:'float'; FieldType:ftFloat),
            (Name:'boolean'; FieldType:ftBoolean),
            (Name:'curency'; FieldType:ftCurrency),
            (Name:'bcd'; FieldType:ftBCD),
            (Name:'datetime'; FieldType:ftDateTime), // must be before 'date'
            (Name:'date'; FieldType:ftDate),
            (Name:'time'; FieldType:ftTime),
            (Name:'varbytes'; FieldType:ftVarBytes),
            (Name:'bytes'; FieldType:ftBytes),
            (Name:'autoinc'; FieldType:ftAutoInc),
            (Name:'fmtmemo'; FieldType:ftFmtMemo),   // must be before memo
            (Name:'memo'; FieldType:ftMemo),
            (Name:'blob'; FieldType:ftBlob),
            (Name:'graphic'; FieldType:ftGraphic),
            (Name:'typedbinary'; FieldType:ftTypedBinary),
            (Name:'fixedchar'; FieldType:ftfixedchar));

{ Default Fields Type when DatabaseType is dtbKexi }
KexiFieldTypes:array[0..12] of TSqlitePassPresetFieldType =
           ((Name:'integer'; Fieldtype:ftInteger),
            (Name:'shortinteger'; FieldType:ftSmallInt),
            (Name:'biginteger'; FieldType:ftLargeint),
            (Name:'byte'; FieldType:ftSmallInt),
            (Name:'double'; FieldType:ftFloat),
            (Name:'float'; FieldType:ftFloat),
            (Name:'boolean'; FieldType:ftBoolean),
            (Name:'datetime'; FieldType:ftDateTime), // must be before 'date'
            (Name:'date'; FieldType:ftDate),
            (Name:'time'; FieldType:ftTime),
            (Name:'text'; FieldType:ftString),
            (Name:'clob'; FieldType:ftMemo),
            (Name:'blob'; FieldType:ftBlob));

{ Default Fields Type when DatabaseType is dbtSqliteExpert }
SqliteExpertFieldTypes:array[0..40] of TSqlitePassPresetFieldType =
           ((Name:'autoinc'; Fieldtype:ftAutoInc),
            (Name:'bigint'; FieldType:ftLargeint),
            (Name:'binary'; FieldType:ftBlob),
            (Name:'blob'; FieldType:ftBlob),
            (Name:'boolean'; FieldType:ftBoolean),
            (Name:'char'; FieldType:ftString),
            (Name:'clob'; FieldType:ftMemo),
            (Name:'currency'; FieldType:ftCurrency),
            (Name:'date'; FieldType:ftDate),
            (Name:'dateText'; FieldType:ftString),
            (Name:'dateTime'; FieldType:ftDateTime),
            (Name:'dec'; FieldType:ftFloat),
            (Name:'decimal'; FieldType:ftFloat),
            (Name:'double'; FieldType:ftFloat),
            (Name:'double precision'; FieldType:ftFloat),
            (Name:'float'; FieldType:ftFloat),
            (Name:'graphic'; FieldType:ftGraphic),
            (Name:'guid'; FieldType:ftMemo), //
            (Name:'int'; FieldType:ftInteger),
            (Name:'int64'; FieldType:ftLargeInt),
            (Name:'integer'; FieldType:ftInteger),
            (Name:'largeint'; FieldType:ftLargeInt),
            (Name:'memo'; FieldType:ftMemo),
            (Name:'nchar'; FieldType:ftString),
            (Name:'money'; FieldType:ftCurrency),
            (Name:'ntext'; FieldType:ftString),
            (Name:'number'; FieldType:ftFloat),
            (Name:'numeric'; FieldType:ftFloat),
            (Name:'nvarchar'; FieldType:ftString),
            (Name:'nvarchar2'; FieldType:ftString),
            (Name:'raw'; FieldType:ftBlob),
            (Name:'real'; FieldType:ftFloat),
            (Name:'smallint'; FieldType:ftInteger),
            (Name:'smallmoney'; FieldType:ftCurrency),
            (Name:'text'; FieldType:ftString),
            (Name:'time'; FieldType:ftTime),
            (Name:'timestamp'; FieldType:ftDateTime),
            (Name:'tinyint'; FieldType:ftInteger),
            (Name:'varbinary'; FieldType:ftBlob),
            (Name:'varchar'; FieldType:ftString),
            (Name:'varchar2'; FieldType:ftString));

{ Default Fields Type when DatabaseType is dbtFPC
  Support for TCustomSqliteDataset Component from Luiz Amrico Pereira Cmara}
FpcFieldTypes:array[0..13] of TSqlitePassPresetFieldType =
           ((Name:'autoinc'; Fieldtype:ftAutoInc),
            (Name:'boolean'; FieldType:ftBoolean),
            (Name:'currency'; FieldType:ftCurrency),
            (Name:'date'; FieldType:ftDate),
            (Name:'dateTime'; FieldType:ftDateTime),
            (Name:'float'; FieldType:ftFloat),
            (Name:'int'; FieldType:ftInteger),
            (Name:'integer'; FieldType:ftInteger),
            (Name:'largeint'; FieldType:ftLargeInt),
            (Name:'numeric'; FieldType:ftFloat),
            (Name:'text'; FieldType:ftMemo),
            (Name:'time'; FieldType:ftTime),
            (Name:'varchar'; FieldType:ftString),
            (Name:'word'; FieldType:ftWord));

    { Loads the requested Default Datatypes Mapping }
    procedure SetTranslationRules(FieldsTypes: array of TSqlitePassPresetFieldType);
    var
    i: Word;
    FieldTypeTranslator: TSqlitePassFieldTypeTranslationRule;
    begin
    ClearAndFreeItems;
    for i := 0 to High(FieldsTypes) do
       begin
       FieldTypeTranslator := TSqlitePassFieldTypeTranslationRule.Create;
       With FieldTypeTranslator do
            begin
            DatatypeName := FieldsTypes[i].Name;
            MappingMode  := mmAll;
            FieldType    := FieldsTypes[i].FieldType;
            end;
       Add(FieldTypeTranslator);
       end;
    end;

begin
Case DatabaseType of
     dbtKexi         : SetTranslationRules(KexiFieldTypes);
     dbtSqliteExpert : SetTranslationRules(SqliteExpertFieldTypes);
     else              SetTranslationRules(DefaultFieldsTypes);
     end;
end;


{ TSqlitePassCustomFieldDefs }

constructor TSqlitePassCustomFieldDefs.Create;
begin
inherited Create;
end;

Destructor TSqlitePassCustomFieldDefs.Destroy;
begin
ClearAndFreeItems;
inherited Destroy;
end;

function TSqlitePassCustomFieldDefs.GetItem(Index: Integer): TSqlitePassCustomFieldDef;
begin
  Result:=TSqlitePassCustomFieldDef(Inherited Items[Index]);
end;

procedure TSqlitePassCustomFieldDefs.SetItem(Index: Integer;
  CustomFieldDef: TSqlitePassCustomFieldDef);
begin
   Put(Index,CustomFieldDef);
end;

function TSqlitePassCustomFieldDefs.Add(CustomFieldDef: TSqlitePassCustomFieldDef): Integer;
begin
   Result:=Inherited Add(CustomFieldDef);
end;

procedure TSqlitePassCustomFieldDefs.ClearAndFreeItems;
var
i: Integer;
begin
{ Clear the internal items }
for i := 0 to Pred(Count)
    do Items[i].Free;
{ Clear the list pointers }
inherited Clear;
end;


{ TSqlitePassTranslator }

{ --- constructor / Destructor --- }
constructor TSqlitePassTranslator.Create(Owner: TSqlitePassDatabase);
begin
Inherited Create;
FDatabase := Owner;
FFieldTypesTranslator := FDatabase.FDatatypeOptions.TranslationRules;
end;

destructor TSqlitePassTranslator.Destroy;
begin
FFieldTypesTranslator := nil;
FDatabase := nil;
Inherited Destroy;
end;


{ Date/Time to String conversion }

function TSqlitePassTranslator._DateToStr(const Date: TDateTime): string;
begin
  DateTimeToString(Result, FDatabase.DataTypeOptions.ShortDateFormat, Date);
end;

function TSqlitePassTranslator._TimeToStr(const Time: TDateTime): string;
begin
  DateTimeToString(Result, LongTimeFormat, Time);
end;

function TSqlitePassTranslator._DateTimeToStr(const DateTime: TDateTime): string;
begin
  DateTimeToString(Result, '', DateTime - 693594);
end;

{ String to Date/Time conversion }

function TSqlitePassTranslator._StrToDate(const S: string): TDateTime;
begin
 Result := StrToDate(S) + 693594;
end;

function TSqlitePassTranslator._StrToTime(const S: string): TDateTime;
begin
 Result := StrToTime(S);
end;

function TSqlitePassTranslator._StrToDateTime(const s: string): TDateTime;
begin
 Result := StrToDateTime(S);
end;


{ Float <-> String conversion utility functions }

function TSqlitePassTranslator._StrToFloat(const s: string): Double;
var
ConvertedStr: String;
begin
ConvertedStr := s;

if convertedStr = '' then
   begin
   Result := 0;
   exit;
   end;

if (FDatabase.DataTypeOptions.DecimalSeparator <> SysUtils.DecimalSeparator)
   then ConvertedStr := StringReplace(S, FDatabase.DataTypeOptions.DecimalSeparator, SysUtils.DecimalSeparator, []);
Result := StrToFloat(ConvertedStr);
end;

function TSqlitePassTranslator._FloatToStr(const Float: Double): string;
begin
  Result := FloatToStr(Float);
end;


{ Boolean <-> String conversion utility functions }
function TSqlitePassTranslator._BooleanToStr(const BoolValue: Boolean): string;
begin
If BoolValue
   then Result := 'True'
   else Result := 'False';
end;

function TSqlitePassTranslator._StrToBoolean(const S: string): Boolean;
var
Str: String;
R: WordBool;
Begin
Str := LowerCase(s);
Result := (Pos('false', Str) = 0) and
          (Pos('no', Str)    = 0) and
          (Pos('off', Str)   = 0) and
          (Pos('0', Str)     = 0);
end;


{ ----- TSqlitePassTranslator ----- GetFieldData Functions }

function TSqlitePassTranslator.GetFieldData(Const StrValue: String; Const DataType: TFieldType; FieldBuffer: Pointer): Boolean;
begin
Result := False;
Try
Case DataType of
      ftBoolean    : Result := StrToBooleanFieldBuffer(StrValue, WordBool(FieldBuffer^));
      ftBytes      : Result := StrToByteFieldBuffer(StrValue, String(FieldBuffer));

      ftWord       ,
      ftSmallint   ,
      ftInteger    ,
      ftAutoInc    : Result := StrToIntFieldBuffer(StrValue, Integer(FieldBuffer^));

      ftBCD        ,
      ftCurrency   ,
      ftFloat      : Result := StrToFloatFieldBuffer(StrValue, Double(FieldBuffer^));
      ftDate       : Result := StrToDateFieldBuffer(StrValue, TDateTimeRec(FieldBuffer^));
      ftTime       : Result := StrToTimeFieldBuffer(StrValue, TDateTimeRec(FieldBuffer^));
      ftDateTime   : Result := StrToDateTimeFieldBuffer(StrValue, TDateTimeRec(FieldBuffer^));
      ftString     : Result := StrToFieldBuffer(StrValue, FieldBuffer);
   End;
Except
end;
end;

Function TSqlitePassTranslator.StrToBooleanFieldBuffer(Const StrValue: String; out Value: WordBool): Boolean;
Begin
 Result := True;
 Try
 Value := _StrToBoolean(StrValue);
 Except
 Result := False;
 End;
End;

function TSqlitePassTranslator.StrToIntFieldBuffer(Const StrValue: String; out Value: Integer): Boolean;
Begin
  Result := True;
  Try
   if StrValue = ''
      then Value := 0
      else Value := StrToIntDef(StrValue, 0);
  Except
   Try
     Value := Round(_StrToFloat(StrValue));
   Except
     Result := False;
   End;
  End;
End;

function TSqlitePassTranslator.StrToLargeIntFieldBuffer(Const StrValue: String; out Value: Int64): Boolean;
Begin
  Result := True;
  Try
   if StrValue = ''
      then Value := 0
      else Value := StrToInt64Def(StrValue, 0);
  Except
   Try
     Value := Round(_StrToFloat(StrValue));
   Except
     Result := False;
   End;
  End;
End;

Function TSqlitePassTranslator.StrToFloatFieldBuffer(Const StrValue: String; out Value: Double): Boolean;
Begin
 Result:=True;
 Try
   if StrValue = ''
      then Value := 0
      else Value := _StrToFloat(StrValue);
 Except
   Result := False;
 End;
End;

Function TSqlitePassTranslator.StrToDateFieldBuffer(Const StrValue: String; out Value: TDateTimeRec): Boolean;
Begin
Result := True;
  Try
    Value.Date := 0;
    if StrValue <> ''
       then Value.Date := Trunc(_StrToDate(StrValue));
    Except
       Result := False;
    End;
end;

function TSqlitePassTranslator.StrToTimeFieldBuffer(Const StrValue: String; out Value: TDateTimeRec): Boolean;
Begin
Result := True;
    Try
    Value.Time := 0;
    if StrValue <> ''
       then Value.Time := DateTimeToTimeStamp(_StrToTime(StrValue)).Time;
    Except
    Result := False;
    End;
end;

Function TSqlitePassTranslator.StrToDateTimeFieldBuffer(StrValue: String; out Value: TDateTimeRec): Boolean;
var
TimePos : Integer;
Begin
Result := True;
    Try
    Value.DateTime := 0;
    if StrValue <> ''
       then Value.DateTime := TimeStampToMsecs(DateTimeToTimeStamp(_StrToDateTime(StrValue)));
    Except
    Result := False;
    End;
end;

function TSqlitePassTranslator.StrToFieldBuffer(Const StrValue: String; out Value: Pointer): Boolean;
Begin
  StrLCopy(PChar(Value),PChar(StrValue),Length(StrValue));
  Result := StrValue <> '';
End;

function TSqlitePassTranslator.StrToStrFieldBuffer(const StrValue: String;
  out Value: String): Boolean;
begin
   Value := StrValue;
   Result := StrValue <> '';
end;

function TSqlitePassTranslator.StrToByteFieldBuffer(const StrValue: String;
  out Value: String): Boolean;
begin
  Result := StrToStrFieldBuffer(StrValue, Value);
end;


{ --- SetFieldData Functions --- }

procedure TSqlitePassTranslator.SetFieldData(Const FieldBuffer: Pointer; Const DataType: TFieldType; Out StrValue: String);
begin
 case DataType of
       ftBytes            : ByteFieldBufferToStr(PChar(FieldBuffer), StrValue);
       ftBoolean          : BooleanFieldBufferToStr(WordBool(FieldBuffer^), StrValue);

       ftWord,
       ftSmallint,
       ftInteger,
       ftAutoInc          : IntFieldBufferToStr(Integer(FieldBuffer^), StrValue);

       ftLargeInt         : LargeIntFieldBufferToStr(Int64(FieldBuffer^), StrValue);

       ftBCD,
       ftCurrency,
       ftFloat            : FloatFieldBufferToStr(Double(FieldBuffer^), StrValue);

       ftDate             : DateFieldBufferToStr(TDateTimeRec(FieldBuffer^), StrValue);
       ftTime             : TimeFieldBufferToStr(TDateTimeRec(FieldBuffer^), StrValue);
       ftDateTime         : DateTimeFieldBufferToStr(TDateTimeRec(FieldBuffer^), StrValue);
       ftString           : FieldBufferToStr(PChar(FieldBuffer), StrValue);
      end;
end;


procedure TSqlitePassTranslator.BooleanFieldBufferToStr(const Value: Boolean; out StrValue: String);
begin
  StrValue := _BooleanToStr(Value);
end;

procedure TSqlitePassTranslator.IntFieldBufferToStr(Const Value: Integer; out StrValue: String);
Begin
  Try
   StrValue := IntToStr(Value);
  Except
   StrValue := '';
  End;
End;

procedure TSqlitePassTranslator.LargeIntFieldBufferToStr(Const Value: Int64; out StrValue: String);
Begin
  Try
   StrValue := IntToStr(Value);
  Except
   StrValue := '';
  End;
End;

procedure TSqlitePassTranslator.FloatFieldBufferToStr(Const Value: Double; out StrValue: String);
Begin
 Try
  StrValue := _FloatToStr(Value);
 Except
  StrValue := '';
 End;
End;

procedure TSqlitePassTranslator.DateFieldBufferToStr(const Value: TDateTimeRec; out StrValue: String);
Begin
    Try
    StrValue := _DateToStr(Value.Date);
    Except
    StrValue := '';
    End;
end;

procedure TSqlitePassTranslator.TimeFieldBufferToStr(const Value: TDateTimeRec; out StrValue: String);
Begin
    Try
    StrValue := _TimeToStr(TimeStampToDateTime(TTimeStamp(Value)));
    Except
    StrValue := '';
    End;
end;

procedure TSqlitePassTranslator.DateTimeFieldBufferToStr(const Value: TDateTimeRec; out StrValue: String);
begin
    Try
    StrValue := _DateTimeToStr(TimeStampToDateTime(MsecsToTimeStamp(Value.DateTime)));
    Except
    StrValue := '';
    End;
end;

procedure TSqlitePassTranslator.FieldBufferToStr(Const Value: PChar; out StrValue: String);
begin
StrValue := StrPas(Value);
end;

procedure TSqlitePassTranslator.ByteFieldBufferToStr(Const Value: PChar; out StrValue: String);
begin
StrValue := StrPas(Value);
end;

{ Binds value to parameters in INSERT - UPDATE - DELETE SQL statement }
function TSqlitePassTranslator.BindValue(const PreparedStmt: pointer; Const DataType: TFieldType; const FieldValue: String; const Idx: Integer): Boolean;
var
Str : String;
Int : Integer;
LargeInt: Int64;
WBool: WordBool;
Dble: Double;
BlobData: TMemoryStream;
begin
Result := False;
if not assigned(PreparedStmt) then exit;

case DataType of
 ftString, ftWideString, ftFixedChar,
 ftDate, ftTime, ftDateTime:
    begin
     Result := StrToStrFieldBuffer(FieldValue, Str);
     if Result
        then FDatabase.Engine.CheckResult(SqliteDbv3_bind_text(PreparedStmt, idx, PChar(Str), Length(Str), SQLITE_TRANSIENT));
    end;

 ftSmallInt, ftInteger, ftWord:
    begin
    Result := StrToIntFieldBuffer(FieldValue, Int);
    If Result
       then FDatabase.Engine.CheckResult(SqliteDbv3_bind_int(PreparedStmt, idx, Int));
    end;

 ftLargeInt:
    begin
    Result := StrToLargeIntFieldBuffer(FieldValue, LargeInt);
    If Result
       then FDatabase.Engine.CheckResult(SqliteDbv3_bind_int64(PreparedStmt, idx, LargeInt));
    end;

 ftBoolean:
   begin
    Result := StrToBooleanFieldBuffer(FieldValue, WBool);
    if Result
       then FDatabase.Engine.CheckResult(SqliteDbv3_bind_int(PreparedStmt, idx, Integer(WBool)))
   end;

 ftBcd, ftFloat, ftCurrency:
    begin
     Result := StrToFloatFieldBuffer(FieldValue, Dble);
     if Result
        then FDatabase.Engine.CheckResult(SqliteDbv3_bind_double(PreparedStmt, idx, Dble));
     end;

 ftBlob, ftGraphic, ftBytes, ftVarBytes, ftMemo, ftFmtMemo:
    begin
        if FieldValue <> '' then
        begin
          Try
          BlobData := TMemoryStream(StrToInt(FieldValue));
          FDatabase.Engine.CheckResult(SqliteDbv3_bind_blob(PreparedStmt,idx,BlobData.Memory,BlobData.Size, SQLITE_TRANSIENT));
          Result := True;
          finally
          BlobData := nil;
          end;
        end;
    end;

 ftUnknown: FDatabase.Engine.CheckResult(SqliteDbv3_bind_null(PreparedStmt,idx));
 else raise Exception.CreateFmt(Msg3011,[ord(DataType)]);
end;

if Not Result
   then FDatabase.Engine.CheckResult(SqliteDbv3_bind_null(PreparedStmt,idx));
end;

(*function TSqlitePassTranslator.CreateTable(const Dataset: TSqlitePassDataset): Boolean;
begin
// TODO : TSqlitePassTranslator.CreateTable - To be implemented
end;*)


{ Database objects information is set from sqlite_master and
 sqlite_temp_master tables. They return basic information :
 Rowid, type, name, tbl_name, rootpage, sql (Create stmt)}

procedure TSqlitePassTranslator.GetDatabaseTableDefs(TableDefs: TSqlitePassTableDefs);
var
i: integer;
Sql: String;
Rec: TSqlitePassMemRecord;
TableDefsRecordset: TSqlitePassRecordset;

  Procedure GetTableDefs(AliasName: String; DatabaseConnection: Pointer);
  var
  i, j: integer;
  FieldDef: TSqlitePassTableFieldDef;
  TableDef: TSqlitePassTableDef;
  begin
  TableDefsRecordset.Open(DatabaseConnection, PChar(Sql));
  { Type, name, tablename, rootpage, sql }
  Rec:=TSqlitePassMemRecord.Create(6);
  Rec.FieldsValues[0] := 'table';
  Rec.FieldsValues[1] := '0';
  Rec.FieldsValues[2] := 'sqlite_master';
  Rec.FieldsValues[3] := 'sqlite_master';
  Rec.FieldsValues[4] := '';
  TableDefsRecordset.Add(Rec);

  for i := 0 to Pred(TableDefsRecordset.count)
   do begin
      TableDef := TSqlitePassTableDef.Create(TableDefs);
      TableDef.FTableName := TableDefsRecordset[i].FieldsValues[2];
      if AliasName <> ''
         then TableDef.FTableFullName := AliasName + '.' + TableDef.FTableName
         else TableDef.FTableFullName := TableDef.FTableName;
      TableDef.FSql := TableDefsRecordset[i].FieldsValues[5];

      { Sets the FieldDefs properties }
      GetDatabaseTableFieldDefs(TableDef);

      { Sets the PrimaryKey info }
      TableDef.FPrimaryKeyField := nil;
      TableDef.FPrimaryKeyFieldNo := -1;
      for j := 0 to Pred(TableDef.FFieldDefs.Count) do
          if TableDef.FFieldDefs[j].PrimaryKey then
             begin
             TableDef.FPrimaryKeyField := TableDef.FFieldDefs[j];
             TableDef.FPrimaryKeyFieldNo := j;
             Break;
             end;

      { No PrimaryKey was found ? then we create one and name it 'Rowid' }
      if TableDef.FPrimaryKeyFieldNo = -1 then
         begin
         FieldDef := TSqlitePassTableFieldDef.Create(TableDef.FFieldDefs);
         With FieldDef do
              begin
              FFieldName := 'Rowid';
              FDisplayName :=  'Rowid';
              FNativeDataTypeCode := 0;
              FNativeDataType := 'integer';
              FDataType := ftInteger;
              FFieldNo := 0;
              FDefaultValue := '';
              FHint := 'Primary Key';
              FNoConstraints := False;
              FAutoInc := True;
              FUnique := True;
              FPrimaryKey := True;
              FForeignKey := False;
              FNotNull := True;
              FNotEmpty := True;
              FIndexed := True;
              end; { With FieldDef }
         TableDef.FFieldDefs.Add(FieldDef);
         TableDef.FPrimaryKeyFieldNo := Pred(TableDef.FieldDefs.Count);
         TableDef.FPrimaryKeyField := FieldDef;
         end;

      TableDefs.Add(TableDef);
      end;
  TableDefsRecordset.Close;
  end;

begin
Sql := 'SELECT Rowid, * FROM sqlite_master WHERE type = "table" ' +
       'UNION ALL SELECT Rowid, * FROM sqlite_temp_master WHERE type = "table" ' +
       'ORDER BY type, tbl_name;';
TableDefs.ClearAndFreeItems;
TableDefsRecordset:= TSqlitePassRecordset.Create(nil);
GetTableDefs('', FDatabase.Engine.Connection);
{ Retrieves TableDefs info for attached databases }
For i := 0 to Pred(FDatabase.Databases.Count) do
    begin
    FDatabase.Databases[i].FEngine.OpenDatabase(FDatabase.Databases[i].Database);
    GetTableDefs(FDatabase.Databases[i].AliasName, FDatabase.Databases[i].FEngine.Connection);
    FDatabase.Databases[i].FEngine.CloseDatabase;
    end;

TableDefsRecordset.Free;
end;

procedure TSqlitePassTranslator.GetDatabaseIndexDefs(IndexDefs: TSqlitePassDatabaseIndexDefs);
var
i: Integer;
Sql, SqlIndexDefs, SqlIndexList, SqlIndexColumns: String;
IndexDefsRecordset: TSqlitePassRecordset;

  Procedure GetIndexDefs(DatabaseConnection: Pointer);
  var
  i, j, SubStringBegin: Integer;
  DirectionStr: String;
  Idx: TSqlitePassIndex;
  Col: TSqlitePassIndexColumn;
  begin
  { We get info on IndexName, DatasetName and SQLCreateStmt }
  IndexDefsRecordset.Open(DatabaseConnection, PChar(SqlIndexDefs));
  for i := 0 to Pred(IndexDefsRecordset.count) do
      begin
      Idx := TSqlitePassDatabaseIndex.Create(FDatabase.IndexDefs);
      Idx.FIndexName := IndexDefsRecordset[i].FieldsValues[2];
      Idx.FDatasetName := IndexDefsRecordset[i].FieldsValues[3];
      Idx.FSql := IndexDefsRecordset[i].FieldsValues[5];
      IndexDefs.Add(Idx);
      end;

  { For each table, we check if the index is unique or not }
  for i := 0 to Pred(FDatabase.TableDefs.Count) do
      begin
      Sql := Format(SqlIndexList, [FDatabase.TableDefs[i].TableName]);
      IndexDefsRecordset.Open(FDatabase.Engine.Connection, PChar(Sql));
      for j := 0 to Pred(IndexDefsRecordset.count) do
          begin
          Idx := IndexDefs.FindIndex(IndexDefsRecordset[j].FieldsValues[1]);
          if Assigned(Idx)
             then Idx.Unique := (IndexDefsRecordset[j].FieldsValues[2] = '1');
          end;
      end;

  { For each index, we get info about fields used and sort order }
  for i := 0 to Pred(IndexDefs.Count) do
      begin
      Sql := Format(SqlIndexColumns, [IndexDefs[i].IndexName]);
      IndexDefsRecordset.Open(FDatabase.Engine.Connection, PChar(Sql));
      for j := 0 to Pred(IndexDefsRecordset.count)
          do begin
          Col := TSqlitePassIndexColumn.Create;
          Col.Position := StrToInt(IndexDefsRecordset[j].FieldsValues[1]);
          Col.ColumnName := IndexDefsRecordset[j].FieldsValues[2];
          SubStringBegin := Pos('"'+Col.ColumnName+'"', IndexDefs[i].Sql);
          If SubStringBegin <> -1
             then begin
             DirectionStr := Copy(IndexDefs[i].Sql, SubStringBegin + 2 + length(Col.ColumnName), 6);
             If Pos('ASC',DirectionStr) <> 0
                then Col.Direction := cidAscending
                else if Pos('DESC',DirectionStr) <> 0
                        then Col.Direction := cidDescending
                        else Col.Direction := cidUnknown;
                end;
          IndexDefs[i].Columns.Add(Col);
          end;
     end;
  IndexDefsRecordset.Close;
  end;

begin
SqlIndexDefs := 'SELECT Rowid, * FROM sqlite_master WHERE type = "index" ' +
                'UNION ALL SELECT Rowid, * FROM sqlite_temp_master WHERE type = "index" ' +
                'ORDER BY type, tbl_name;';
SqlIndexList := 'PRAGMA index_list ("%s");';
SqlIndexColumns := 'PRAGMA index_info ("%s");';
IndexDefs.ClearAndFreeItems;
IndexDefsRecordset:= TSqlitePassRecordset.Create(nil);
GetIndexDefs(FDatabase.Engine.Connection);
{ Retrieves IndexDefs info for attached databases }
For i := 0 to Pred(FDatabase.Databases.Count) do
    begin
    FDatabase.Databases[i].FEngine.OpenDatabase(FDatabase.Databases[i].Database);
    GetIndexDefs(FDatabase.Databases[i].FEngine.Connection);
    FDatabase.Databases[i].FEngine.CloseDatabase;
    end;
IndexDefsRecordset.Free;
end;

function TSqlitePassTranslator.IsSystemTable(TableName: String): Boolean;
begin
Result :=    (CompareText(TableName, 'SqlitePass__DbSettings') = 0)
          or (CompareText(TableName, 'sqlite_master') = 0);
end;

procedure TSqlitePassTranslator.GetDatabaseTableFieldDefs(TableDef: TSqlitePassTableDef);
var
aFieldDef: TSqlitePassTableFieldDef;
i: integer;
FieldDefsRecordset: TSqlitePassRecordset;
Sql: String;

 function GetFieldDefSize(FieldDefTypeStr: String): Integer;
  var
  StartPos, EndPos: Word;
  begin
  Result := 1;
  Try
  StartPos:=pos('(',FieldDefTypeStr);
  if StartPos > 0
  then begin
       EndPos:=pos(')',FieldDefTypeStr);
       if EndPos > StartPos
          then result := Trunc(StrToFloat(Copy(FieldDefTypeStr,StartPos+1,EndPos-StartPos-1)));
       end;
  Except
  Result := 0;
  end;
  end;

begin
TableDef.FieldDefs.Clear;
Sql := 'PRAGMA table_info ("' + TableDef.TableName +'");';

FieldDefsRecordset := TSqlitePassRecordset.Create(nil);
FieldDefsRecordset.Open(FDatabase.Engine.Connection, PChar(Sql));

for i := 0 to Pred(FieldDefsRecordset.count) do
    begin
    aFieldDef := TSqlitePassTableFieldDef.Create(TableDef.FieldDefs);
    With aFieldDef do
         begin
         With FieldDefsRecordset[i] do
              begin
              FFieldName := FieldsValues[1];
              FNativeDataType := FieldsValues[2];
              FNotNull := FieldsValues[3] <> '0';
              FDefaultValue := FieldsValues[4];
              FPrimaryKey := FieldsValues[5] = '1';
              FDataType := FieldTypesTranslator.ConvertNativeFieldTypeToPascalFieldType(FNativeDataType);
              { Now we have to set the field size and precision }
              Case FDataType of
                   ftString : FSize := GetFieldDefSize(FNativeDataType);
                   ftBCD    : FPrecision := 2; // TODO Rework
                   end;
              end;
         end;
    TableDef.FieldDefs.Add(aFieldDef);
    end;
FieldDefsRecordset.Free;
end;

procedure TSqlitePassTranslator.GetDatabaseQueryDefs(QueryDefs: TSqlitePassQueryDefs);
begin
QueryDefs.ClearAndFreeItems;
{ QueryDefs are not stored by sqlite engine, so there is nothing to do
  Implementation must be done in TSqlitePassTranslator descendants }
end;

procedure TSqlitePassTranslator.GetDatabaseViews(Views: TSqlitePassViews);
var
i: Integer;
Sql: String;
ViewsRecordset: TSqlitePassRecordset;

  Procedure GetViews(DatabaseConnection: Pointer);
  var
  CreateStmtStart, i: Integer;
  View: TSqlitePassView;
  begin
  ViewsRecordset.Open(DatabaseConnection, PChar(Sql));

  for i := 0 to Pred(ViewsRecordset.count) do
      begin
      View := TSqlitePassView.Create(Views);
      View.FViewName := ViewsRecordset[i].FieldsValues[3];
      View.FSqlCreateStmt := ViewsRecordset[i].FieldsValues[5];
      CreateStmtStart := Pos(kwSelect + ' ', View.FSqlCreateStmt);
      if CreateStmtStart <> 0
         then View.FSqlSelectStmt := Copy(View.FSqlCreateStmt, CreateStmtStart, Length(View.FSqlCreateStmt));
      Views.Add(View);
      end;

  ViewsRecordset.Close;
  end;


begin
Sql := 'SELECT Rowid, * FROM sqlite_master WHERE type = "view" ' +
       'UNION ALL SELECT Rowid, * FROM sqlite_temp_master WHERE type = "view" ' +
       'ORDER BY type, tbl_name;';
Views.ClearAndFreeItems;
ViewsRecordset := TSqlitePassRecordset.Create(nil);
GetViews(FDatabase.Engine.Connection);
{ Retrieves Views info for attached databases }
For i := 0 to Pred(FDatabase.Databases.Count) do
    begin
    FDatabase.Databases[i].FEngine.OpenDatabase(FDatabase.Databases[i].Database);
    GetViews(FDatabase.Databases[i].FEngine.Connection);
    FDatabase.Databases[i].FEngine.CloseDatabase;
    end;
ViewsRecordset.Free;
end;

procedure TSqlitePassTranslator.GetDatabaseTriggers(Triggers: TSqlitePassTriggers);
var
i: Integer;
Sql: String;
TriggersRecordset: TSqlitePassRecordset;

  Procedure GetTriggers(DatabaseConnection: Pointer);
  var
  i: integer;
  Trigger: TSqlitePassTrigger;

  begin
  TriggersRecordset.Open(DatabaseConnection, PChar(Sql));

  for i := 0 to Pred(TriggersRecordset.count)
   do begin
      Trigger := TSqlitePassTrigger.Create(Triggers);
      Trigger.FTriggerName := TriggersRecordset[i].FieldsValues[2];
      Trigger.FSQL := TriggersRecordset[i].FieldsValues[3];
      //Trigger.FAttached := False;  // TODO could be True too...
      Triggers.Add(Trigger);
      end;

  TriggersRecordset.Close;
  end;

begin
Sql := 'SELECT Rowid, * FROM sqlite_master WHERE type = "trigger" ' +
       'UNION ALL SELECT Rowid, * FROM sqlite_temp_master WHERE type = "trigger" ' +
       'ORDER BY type, tbl_name;';
Triggers.ClearAndFreeItems;
TriggersRecordset := TSqlitePassRecordset.Create(nil);
GetTriggers(FDatabase.Engine.Connection);
{ Retrieves Views info for attached databases }
For i := 0 to Pred(FDatabase.Databases.Count) do
    begin
    FDatabase.Databases[i].FEngine.OpenDatabase(FDatabase.Databases[i].Database);
    GetTriggers(FDatabase.Databases[i].FEngine.Connection);
    FDatabase.Databases[i].FEngine.CloseDatabase;
    end;
TriggersRecordset.Free;
end;


{ TSqlitePassTranslator_Kexi }

{ Date/Time to String conversion }

function TSqlitePassTranslator_Kexi._DateToStr(const Date: TDateTime): string;
var
Year, Month, Day: Word;
begin
  DecodeDate(Date - 693594, Year, Month, Day);
  Result := Format('%0.4d',[Year])
         +  FDatabase.DatatypeOptions.DateSeparator
         +  Format('%0.2d',[Month])
         +  FDatabase.DatatypeOptions.DateSeparator
         +  Format('%0.2d',[Day]);
end;

function TSqlitePassTranslator_Kexi._TimeToStr(const Time: TDateTime): string;
var
Hour, Min, Sec, MSec: Word;
begin
  DecodeTime(Time, Hour, Min, Sec, MSec);
  Result := Format('%0.2d',[Hour])
         +  FDatabase.DatatypeOptions.TimeSeparator
         +  Format('%0.2d',[Min])
         +  FDatabase.DatatypeOptions.TimeSeparator
         +  Format('%0.2d',[Sec]);
end;

function TSqlitePassTranslator_Kexi._DateTimeToStr(const DateTime: TDateTime): string;
begin
  Result := _DateToStr(DateTime + 693594) + 'T' + _TimeToStr(DateTime);
end;

{ String to Date/Time conversion }

function TSqlitePassTranslator_Kexi._StrToDate(const S: string): TDateTime;
var
Year, Month, Day: Word;
begin
 Year := StrToIntDef(System.Copy(S,1,4),1);
 Month := StrToIntDef(System.Copy(S,6,2),1);
 Day := StrToIntDef(System.Copy(S,9,2),1);
 Result := EncodeDate(Year, Month, Day) + 693594;
end;

function TSqlitePassTranslator_Kexi._StrToTime(const S: string): TDateTime;
var
Hour, Min, Sec: Word;
begin
 Try
 Hour := StrToIntDef(System.Copy(S,1,2),1);
 Min := StrToIntDef(System.Copy(S,4,2),1);
 Sec := StrToIntDef(System.Copy(S,7,2),1);
 Result := EncodeTime(Hour, Min, Sec, 0);
 Except
 Result := 0;
 end;
end;

function TSqlitePassTranslator_Kexi._StrToDateTime(const S: string): TDateTime;
var
DateStr, TimeStr: String;
Date, Time: TDateTime;
begin
DateStr := System.Copy(S, 1, 10);
TimeStr := System.Copy(S, 12, 8);
Date := _StrToDate(DateStr) - 693594;
Time := _StrToTime(TimeStr);
if Date < 0
   then Result := trunc(Date) - frac(Time)
   else Result := trunc(Date) + frac(Time);
end;

{ Boolean <-> String conversion utility functions }

function TSqlitePassTranslator_Kexi._BooleanToStr(const BoolValue: Boolean): string;
begin
If BoolValue
   then Result := '1'
   else Result := '0';
end;

function TSqlitePassTranslator_Kexi.GetNativeFieldTypeName(
  FieldType: Word): String;
begin
Case FieldType of
     // TODO - Result := ftBytes doesn't return correct data - Changed to ftSmallInt;
     1: Result := 'byte';
     2: Result := 'shortinteger';
     3: Result := 'integer';
     // TODO - LargeInt doesn't seem to be supported by kexi binding (Check autoInc field) 4: Result := ftLargeint;
     4: Result := 'biginteger';
     5: Result := 'boolean';
     6: Result := 'date';
     7: Result := 'datetime';
     8: Result := 'time';
     9: Result := 'float';
     10: Result := 'float';
     11: Result := 'text';
     12: Result := 'clob';
     13: Result := 'blob';// ftGraphic;
     else Result := 'unknown';
     end;
end;

function TSqlitePassTranslator_Kexi.ConvertNativeFieldTypeToPascalFieldType(
  FieldType: Word): TFieldType;
begin
{ Translated from Kexi project : field.h }
Case FieldType of
     // Byte = 1
     // TODO - Result := ftBytes doesn't return correct data - Changed to ftSmallInt;
     1: Result := ftSmallInt;
     // ShortInteger = 2
     2: Result := ftSmallInt;
     // Integer = 3
     3: Result := ftinteger;
     // BigInteger = 4
     // TODO - LargeInt doesn't seem to be supported by kexi binding (Check autoInc field) 4: Result := ftLargeint;
     4: Result := ftInteger;
     // Boolean = 5
     5: Result := ftBoolean;
     // Date = 6
     6: Result := ftDate;
     // DateTime = 7
     7: Result := ftDateTime;
     // Time = 8
     8: Result := ftTime;
     // Float = 9
     9: Result := ftFloat;
     // Double = 10
     10: Result := ftFloat;
     // Text = 11
     11: Result := ftString;
     // LongText = 12
     12: Result := ftMemo;
     // BLOB = 13
     13: Result := ftBlob;// ftGraphic;
     else Result := ftunknown;
     end;
end;

procedure TSqlitePassTranslator_Kexi.GetDatabaseTableFieldDefs(TableDef: TSqlitePassTableDef);
const
SqlConst = 'SELECT t1.o_id,t1.o_name,t1.o_type,t2.f_type,t2.f_name,t2.f_length,'
    + 't2.f_precision,t2.f_constraints,t2.f_options,t2.f_default,t2.f_order,'
    + 't2.f_caption,t2.f_help '
    + 'FROM kexi__objects t1,kexi__fields t2 '
    + 'WHERE t1.o_id=t2.t_id AND t1.o_name = "%s";' ;
var
i: integer;
FieldConstraints: TSqlitePassBitArray;
aFieldDef: TSqlitePassTableFieldDef;
Sql: String;
FieldDefsRecordset: TSqlitePassRecordset;

begin
TableDef.FieldDefs.Clear;
FieldDefsRecordset := TSqlitePassRecordset.Create(nil);
Sql := Format(SqlConst,[TableDef.TableName]);
FieldDefsRecordset.Open(FDatabase.Engine.Connection, PChar(Sql));
FieldConstraints := TSqlitePassBitArray.Create;
FieldConstraints.Items := KexiFieldConstraints;
// TODO - Handle System Tables Kexi__...

for i := 0 to Pred(FieldDefsRecordset.count) do
    begin
    aFieldDef := TSqlitePassTableFieldDef.Create(TableDef.FieldDefs);
    With aFieldDef do
         begin
         With FieldDefsRecordset[i] do
              begin
              FFieldName := FieldsValues[4];
              FDisplayName :=  FieldsValues[11];
              FNativeDataTypeCode := StrToInt(FieldsValues[3]);
              FNativeDataType := GetNativeFieldTypeName(FNativeDataTypeCode);

              Case FDatabase.DataTypeOptions.DetectionMode of
                   dmTypeName  : FDataType := FDatabase.DataTypeOptions.TranslationRules.ConvertNativeFieldTypeToPascalFieldType(FNativeDataType);
                   dmDbSpecific: begin
                                 FDataType := ConvertNativeFieldTypeToPascalFieldType(FNativeDataTypeCode);
                                 if FDataType = ftUnknown
                                    then FDataType := FieldTypesTranslator.ConvertNativeFieldTypeToPascalFieldType(FNativeDataType);
                                 end;
                   dmCustom    : begin
                                 If Assigned(FDatabase.OnDataTypeConversion) then
                                    begin
                                    FDataType := FDatabase.DataTypeOptions.TranslationRules.ConvertNativeFieldTypeToPascalFieldType(FNativeDataType);
                                    if FDataType = ftUnknown
                                       then FDataType := ConvertNativeFieldTypeToPascalFieldType(FNativeDataTypeCode);
                                    FDatabase.OnDataTypeConversion(FDatabase, TableDef, FNativeDataType, FDataType);
                                    end;
                                 end;
                   dmForceStr  : FDataType := ftString;
                   end;

              FFieldNo := StrToInt(FieldsValues[10]);
              FDefaultValue := FieldsValues[9];
              FHint := FieldsValues[12];

              { Now we have to set the field size and precision }
              Case AFieldDef.DataType of
                   ftString : FSize := StrToInt(FieldsValues[5]);
                   ftBCD    : FPrecision := StrToInt(FieldsValues[6]);
                   end;

              { Now we have to set the field constraints }
              FieldConstraints.Value := StrToInt(FieldsValues[7]);
              end; {With FieldDefsRecordset[i] }
          FNoConstraints := FieldConstraints.IsTrue('NoConstraints');
          FAutoInc := FieldConstraints.IsTrue('AutoInc');
          FUnique := FieldConstraints.IsTrue('Unique');
          FPrimaryKey := FieldConstraints.IsTrue('PrimaryKey');
          FForeignKey := FieldConstraints.IsTrue('ForeignKey');
          FNotNull := FieldConstraints.IsTrue('NotNull');
          FNotEmpty := FieldConstraints.IsTrue('NotEmpty');
          FIndexed := FieldConstraints.IsTrue('Indexed');
          end; { With aFieldDef }
    TableDef.FieldDefs.Add(aFieldDef);
    end; { for }

FieldConstraints.Free;
FieldDefsRecordset.Close;
FieldDefsRecordset.Free;

{ If we don't find any fieldefs (system tables fielddefs are not returned by kexi)
  we try the classic method... }
if TableDef.FieldDefs.Count = 0
   then inherited GetDatabaseTableFieldDefs(TableDef);
end;

procedure TSqlitePassTranslator_Kexi.GetDatabaseQueryDefs(QueryDefs: TSqlitePassQueryDefs);
var
i: Integer;
Sql: String;
QueryDefsRecordset: TSqlitePassRecordset;

  Procedure GetQueryDefs(AliasName: String; DatabaseConnection: Pointer);
  var
  i: integer;
  QueryDef: TSqlitePassQueryDef;
  begin
  QueryDefsRecordset.Open(DatabaseConnection, PChar(Sql));
  for i := 0 to Pred(QueryDefsRecordset.count)
   do begin
      QueryDef := TSqlitePassQueryDef.Create(QueryDefs);
      QueryDef.FQueryName := QueryDefsRecordset[i].FieldsValues[3];
      if AliasName <> ''
         then QueryDef.FQueryFullName := AliasName + '.' + QueryDef.FQueryName
         else QueryDef.FQueryFullName := QueryDef.FQueryName;
      QueryDef.FSql := QueryDefsRecordset[i].FieldsValues[4];
      QueryDefs.Add(QueryDef);
      end;
  QueryDefsRecordset.Close;
  end;

begin
Sql := 'SELECT K1.Rowid, K2.rowid, K1.o_type,K1.o_name,K2.o_data '+
       'FROM kexi__objects K1, kexi__objectdata K2 '+
       'WHERE (K1.o_id=K2.o_id) AND (K1.o_type=2) AND (K2.o_sub_id="sql");';
QueryDefs.ClearAndFreeItems;

QueryDefsRecordset := TSqlitePassRecordset.Create(nil);
GetQueryDefs('', FDatabase.Engine.Connection);

{ Retrieves TableDefs info for attached databases }
For i := 0 to Pred(FDatabase.Databases.Count) do
    begin
    FDatabase.Databases[i].FEngine.OpenDatabase(FDatabase.Databases[i].Database);
    GetQueryDefs(FDatabase.Databases[i].AliasName, FDatabase.Databases[i].FEngine.Connection);
    FDatabase.Databases[i].FEngine.CloseDatabase;
    end;

QueryDefsRecordset.Free;
end;

function TSqlitePassTranslator_Kexi.IsSystemTable(
  TableName: String): Boolean;
var
i: Integer;
begin
for i := 0 to High(KexiSystemTables) do
    begin
    Result :=  (CompareText(TableName, KexiSystemTables[i]) = 0);
    if Result then break;
    end;
If Not Result
   then Result := Inherited IsSystemTable(TableName);//((CompareTableName = SqlitePassSystemTable)
end;

{ SQLiteExpert }
{ TSqlitePassTranslator_Kexi }

{ Date/Time to String conversion }

function TSqlitePassTranslator_sqliteExpert._DateToStr(const Date: TDateTime): string;
var
Year, Month, Day: Word;
begin
  DecodeDate(Date - 693594, Year, Month, Day);
  Result := Format('%0.4d',[Year])
         +  FDatabase.DatatypeOptions.DateSeparator
         +  Format('%0.2d',[Month])
         +  FDatabase.DatatypeOptions.DateSeparator
         +  Format('%0.2d',[Day]);
end;

function TSqlitePassTranslator_sqliteExpert._TimeToStr(const Time: TDateTime): string;
var
Hour, Min, Sec, MSec: Word;
begin
  DecodeTime(Time, Hour, Min, Sec, MSec);
  Result := Format('%0.2d',[Hour])
         +  FDatabase.DatatypeOptions.TimeSeparator
         +  Format('%0.2d',[Min])
         +  FDatabase.DatatypeOptions.TimeSeparator
         +  Format('%0.2d',[Sec]);
end;

function TSqlitePassTranslator_sqliteExpert._DateTimeToStr(const DateTime: TDateTime): string;
begin
  Result := _DateToStr(DateTime + 693594) + 'T' + _TimeToStr(DateTime);
end;

{ String to Date/Time conversion }

function TSqlitePassTranslator_sqliteExpert._StrToDate(const S: string): TDateTime;
var
Year, Month, Day: Word;
begin
 Year := StrToIntDef(System.Copy(S,1,4),1);
 Month := StrToIntDef(System.Copy(S,6,2),1);
 Day := StrToIntDef(System.Copy(S,9,2),1);
 Result := EncodeDate(Year, Month, Day) + 693594;
end;

function TSqlitePassTranslator_sqliteExpert._StrToTime(const S: string): TDateTime;
var
Hour, Min, Sec: Word;
begin
 Try
 Hour := StrToIntDef(System.Copy(S,1,2),1);
 Min := StrToIntDef(System.Copy(S,4,2),1);
 Sec := StrToIntDef(System.Copy(S,7,2),1);
 Result := EncodeTime(Hour, Min, Sec, 0);
 Except
 Result := 0;
 end;
end;

function TSqlitePassTranslator_sqliteExpert._StrToDateTime(const S: string): TDateTime;
var
DateStr, TimeStr: String;
Date, Time: TDateTime;
begin
DateStr := System.Copy(S, 1, 10);
TimeStr := System.Copy(S, 12, 8);
Date := _StrToDate(DateStr) - 693594;
Time := _StrToTime(TimeStr);
if Date < 0
   then Result := trunc(Date) - frac(Time)
   else Result := trunc(Date) + frac(Time);
end;

{ Boolean <-> String conversion utility functions }

function TSqlitePassTranslator_sqliteExpert._BooleanToStr(const BoolValue: Boolean): string;
begin
If BoolValue
   then Result := '1'
   else Result := '0';
end;


{ TSqlitePassDatabaseDataTypeOptions }

constructor TSqlitePassDatabaseDataTypeOptions.Create(
  Database: TSqlitePassDatabase);
begin
  Inherited Create;
  FDatabase := Database;
  FDateSeparator := SysUtils.DateSeparator;
  FDecimalSeparator := SysUtils.DecimalSeparator;
  FLongTimeFormat := SysUtils.LongTimeFormat;
  FShortDateFormat := SysUtils.ShortDateFormat;
  FTimeSeparator := SysUtils.TimeSeparator;
  FLoadOptions := [soProperties, soTranslationRules, soCustomFieldDefs];
  FSaveOptions := [soProperties, soTranslationRules, soCustomFieldDefs];
  FFieldTypesTranslationRules := TSqlitePassFieldTypesTranslationRules.Create(Self);;
  FCustomFieldDefs := TSqlitePassCustomFieldDefs.Create;
end;

destructor TSqlitePassDatabaseDataTypeOptions.Destroy;
begin
  FCustomFieldDefs.Free;
  FFieldTypesTranslationRules.Free;
  Inherited Destroy;
end;

function TSqlitePassDatabaseDataTypeOptions.GetOwner: TPersistent;
begin
  Result := FDatabase;
end;

procedure TSqlitePassDatabaseDataTypeOptions.LoadFromDatabase(_LoadOptions: TsqlitePassDataTypeStorageOptions = []);
Const
Sql = 'SELECT * FROM "SqlitePass__DbSettings";';
var
i: Integer;
DbSettingsRecordset: TSqlitePassRecordset;
StorageInfoStrings, MainPropertiesStrings, TranslationRulesStrings, CustomFieldDefsStrings: TStringList;
TranslationRule: TSqlitePassFieldTypeTranslationRule;
CustomFieldDef: TSqlitePassCustomFieldDef;
LoadingError: Boolean;
begin
If FDatabase.FState = dbsInactive then Exit;
LoadingError := False;
Try {1}
  If _LoadOptions = []
     then _LoadOptions := FLoadOptions;

  StorageInfoStrings      := TStringList.Create;
  MainPropertiesStrings   := TStringList.Create;
  TranslationRulesStrings := TStringList.Create;
  CustomFieldDefsStrings  := TStringList.Create;

  Try {2}

     DbSettingsRecordset:= TSqlitePassRecordset.Create(nil);
     Try {3}
     { Loads Options from database }
     if DbSettingsRecordset.Open(FDatabase.Engine.Connection, PChar(Sql))
        And (DbSettingsRecordset.Count > 0) then
        begin
        StorageInfoStrings.SetText(PChar(DbSettingsRecordset[0].FieldsValues[0]));
        MainPropertiesStrings.SetText(PChar(DbSettingsRecordset[0].FieldsValues[1]));
        TranslationRulesStrings.SetText(PChar(DbSettingsRecordset[0].FieldsValues[2]));
        CustomFieldDefsStrings.SetText(PChar(DbSettingsRecordset[0].FieldsValues[3]));
        end
        else LoadingError := True;
     finally
       DbSettingsRecordset.Close;
       DbSettingsRecordset.Free;
     end; { Try 3}

   { Table SqlitePass__DbSettings doesn't exist...}
   if LoadingError then exit;

   { Storage Version }
   FStorageVersion := StorageInfoStrings[1];

   { Main Properties }
   if soProperties in _LoadOptions then
      begin
      i := 1;
      While i < Pred(MainPropertiesStrings.Count) do
            begin
            DetectionMode := TSqlitePassDataTypeDetectionMode(StrToInt(MainPropertiesStrings[i]));
            Inc(i,2);
            DateSeparator := MainPropertiesStrings[i][1];
            Inc(i,2);
            DateStorage   := TSqlitePassDateStorage(StrToInt(MainPropertiesStrings[i]));
            Inc(i,2);
            DecimalSeparator := MainPropertiesStrings[i][1];
            Inc(i,2);
            DefaultFieldType := StringToFieldType(MainPropertiesStrings[i]);
            Inc(i,2);
            LongTimeFormat := MainPropertiesStrings[i];
            Inc(i,2);
            ShortTimeFormat := MainPropertiesStrings[i];
            Inc(i,2);
            TimeSeparator := MainPropertiesStrings[i][1];
            Inc(i,2);
            BooleanExtension := MainPropertiesStrings[i];
            end;
       end;


   { TranslationRules }
   if soTranslationRules in _LoadOptions then
      begin
      TranslationRules.ClearAndFreeItems;
      i := 1;
      While i < Pred(TranslationRulesStrings.Count) do
            begin
            TranslationRule := TSqlitePassFieldTypeTranslationRule.Create;
            TranslationRule.DatatypeName := TranslationRulesStrings[i];
            Inc(i,2);
            TranslationRule.MappingMode :=  TSqlitePassDataTypeMappingMode(StrToInt(TranslationRulesStrings[i]));
            Inc(i,2);
            TranslationRule.FieldType := StringToFieldType(TranslationRulesStrings[i]);
            TranslationRules.Add(TranslationRule);
            Inc(i,2);
            end;
       end;

   { CustomFieldDefs }
   if soCustomFieldDefs in _LoadOptions then
      begin
      CustomFieldDefs.ClearAndFreeItems;
      i := 1;
      While i < Pred(CustomFieldDefsStrings.Count) do
            begin
            CustomFieldDef := TSqlitePassCustomFieldDef.Create;
            CustomFieldDef.TableName := CustomFieldDefsStrings[i];
            Inc(i,2);
            CustomFieldDef.FieldName := CustomFieldDefsStrings[i];
            Inc(i,2);
            CustomFieldDef.FieldType := StringToFieldType(CustomFieldDefsStrings[i]);
            Inc(i,2);
            CustomFieldDef.FieldSize := StrToInt(CustomFieldDefsStrings[i]);
            Inc(i,2);
            CustomFieldDef.FieldPrecision := StrToInt(CustomFieldDefsStrings[i]);
            CustomFieldDefs.Add(CustomFieldDef);
            Inc(i,2);
            end;
      end;
   except
      if MessageDlg(Msg1044, mtWarning, [mbYes, mbNo], 0) = mrYes
         then FDatabase.Engine.Exec('Drop table "SqlitePass__DbSettings";');
   end; {Try 2 }

 finally
 StorageInfoStrings.Free;
 MainPropertiesStrings.Free;
 TranslationRulesStrings.Free;
 CustomFieldDefsStrings.Free;
 end; {Try 1}
end;


procedure TSqlitePassDatabaseDataTypeOptions.SaveToDatabase(_SaveOptions: TsqlitePassDataTypeStorageOptions = []);
Const
Sql = 'CREATE TABLE IF NOT EXISTS "SqlitePass__DbSettings" ("StorageVersion" CBLOB, "MainProperties" CBLOB, "TranslationRules" CBLOB, "CustomFieldDefs" CBLOB);';
var
i: Integer;
IsConnected: Boolean;
DbSettingsDataset: TSqlitePassDataset;
DbSettingTableDef: TSqlitePassTableDef;
DataFieldDef: TSqlitePassTableFieldDef;
Lines: TStringList;
begin
If Assigned(FDatabase) and Assigned(FDatabase.FEngine) then
   begin
   try
     if _SaveOptions = []
        then _SaveOptions := FSaveOPtions;

     IsConnected := FDatabase.Connected;
     FDatabase.Open;

     { Creates the table if it doesn't already exist }
     DbSettingTableDef := FDatabase.FTableDefs.FindTable('SqlitePass__DbSettings');
     if DbSettingTableDef = nil then
        begin
        FDatabase.FEngine.Exec(Sql);
        FDatabase.TableDefs.Refresh;
        DbSettingTableDef := FDatabase.FTableDefs.FindTable('SqlitePass__DbSettings')
        end;

     { Enforces fields fieldTypes to ftMemo whatever Databasetype is used }
     For i := 0 to Pred(DbSettingTableDef.FieldDefs.Count) do
         if Not DbSettingTableDef.FieldDefs[i].PrimaryKey
            then DbSettingTableDef.FieldDefs[i].DataType := ftMemo;

     { Saves Options to database }
     DbSettingsDataset:= TSqlitePassDataset.Create(nil);
     DbSettingsDataset.Database := FDatabase;
     
     With DbSettingsDataset do
          begin
          DatasetName := 'SqlitePass__DbSettings';
          Open;
          First;
          Edit;
          end;

     Lines := TStringList.Create;

     { Save Storage Version }
     With Lines do
          begin
          Add('ComponentVersion');Add(FDatabase.VersionInfo.Component);
          end;
     DbSettingsDataset.FieldByName('StorageVersion').Assign(Lines);

     { Main Properties }
     if soProperties in _SaveOptions then
        begin
        With Lines do
             begin
             Clear;
             Add('DetectionMode'); Add(IntToStr(Ord(DetectionMode)));
             Add('DateSeparator'); Add(DateSeparator);
             Add('DateStorage'); Add(IntToStr(Ord(DateStorage)));
             Add('DecimalSeparator'); Add(DecimalSeparator);
             Add('DefaultFieldType'); Add(FieldTypeToString(DefaultFieldType));
             Add('LongTimeFormat'); Add(LongTimeFormat);
             Add('ShortTimeFormat'); Add(ShortTimeFormat);
             Add('TimeSeparator'); Add(TimeSeparator);
             Add('BooleanExtension'); Add(BooleanExtension);
             DbSettingsDataset.FieldByName('MainProperties').Assign(Lines);
             end;
        end;

     { TranslationRules }
     if soTranslationRules in _SaveOptions then
        begin
          With Lines do
          begin
          Clear;
          for i := 0 to Pred(TranslationRules.Count) do
              begin
              Add('DatatypeName');
              Add(TranslationRules[i].DatatypeName);
              Add('MappingMode');
              Add(IntToStr(Ord(TranslationRules[i].MappingMode)));
              Add('FieldType');
              Add(FieldTypeToString(TranslationRules[i].FieldType));
              end;
          end;
        DbSettingsDataset.FieldByName('TranslationRules').Assign(Lines);
        end;

     { Custom FieldDefs }
      if soCustomFieldDefs in _SaveOptions then
        begin
          With Lines do
          begin
          Clear;
          for i := 0 to Pred(CustomFieldDefs.Count) do
              begin
              Add('TableName'); Add(CustomFieldDefs[i].TableName);
              Add('FieldName'); Add(CustomFieldDefs[i].FieldName);
              Add('FieldType'); Add(FieldTypeToString(CustomFieldDefs[i].FieldType));
              Add('FieldSize'); Add(IntToStr(CustomFieldDefs[i].FieldSize));
              Add('FieldPrecision'); Add(IntToStr(CustomFieldDefs[i].FieldPrecision));
              end;
          end;
        DbSettingsDataset.FieldByName('CustomFieldDefs').Assign(Lines);
        end;

   finally
     DbSettingsDataset.Post;
     DbSettingsDataset.Free;
     Lines.Free;
     FDatabase.Connected := IsConnected;
   end;
   end;
end;

procedure TSqlitePassDatabaseDataTypeOptions.DefineProperties(Filer: TFiler);
begin
{ Property are prefixed with 'p' to have a different name of the original
  property otherwise, DefineProperty is not called by the filer }
  Filer.DefineProperty('pCustomFieldDefs', ReadCustomFieldDefs, WriteCustomFieldDefs, True);
  Filer.DefineProperty('pTranslationsRules', ReadTranslationRules, WriteTranslationRules, True);
end;

procedure TSqlitePassDatabaseDataTypeOptions.ReadTranslationRules(Reader: TReader);
var
TranslationRule: TSqlitePassFieldTypeTranslationRule;
begin
FFieldTypesTranslationRules.ClearAndFreeItems;
Reader.ReadListBegin;
while not(Reader.EndofList) do
      begin
      TranslationRule := TSqlitePassFieldTypeTranslationRule.Create;
      With TranslationRule do
           begin
           DatatypeName := Reader.ReadString;
           MappingMode := TSqlitePassDataTypeMappingMode(Reader.ReadInteger);
           FieldType := StringToFieldType(Reader.ReadString);
           end;
      FFieldTypesTranslationRules.Add(TranslationRule);
      end;
Reader.ReadListEnd;
end;

procedure TSqlitePassDatabaseDataTypeOptions.WriteTranslationRules(Writer: TWriter);
var
  i: Integer;
begin
  With Writer do
       begin
       WriteListBegin;
       for i := 0 to Pred(FFieldTypesTranslationRules.Count) do
          begin
          WriteString(FFieldTypesTranslationRules[i].DatatypeName);
          WriteInteger(Ord(FFieldTypesTranslationRules[i].MappingMode));
          WriteString(FieldTypeToString(FFieldTypesTranslationRules[i].FieldType));
          end;
       WriteListEnd;
       end;
end;

procedure TSqlitePassDatabaseDataTypeOptions.ReadCustomFieldDefs(Reader: TReader);
var
CustomFieldDef: TSqlitePassCustomFieldDef;
begin
FCustomFieldDefs.ClearAndFreeItems;
Reader.ReadListBegin;
while not(Reader.EndofList) do
      begin
      CustomFieldDef := TSqlitePassCustomFieldDef.Create;
      With CustomFieldDef do
           begin
           TableName := Reader.ReadString;
           FieldName := Reader.ReadString;
           FieldType := StringToFieldType(Reader.ReadString);
           FieldSize := Reader.ReadInteger;
           FieldPrecision := Reader.ReadInteger;
           end;
      FCustomFieldDefs.Add(CustomFieldDef);
      end;
Reader.ReadListEnd;
end;

procedure TSqlitePassDatabaseDataTypeOptions.WriteCustomFieldDefs(Writer: TWriter);
var
  i: Integer;
begin
  With Writer do
       begin
       WriteListBegin;
       for i := 0 to Pred(FCustomFieldDefs.Count) do
          begin
          WriteString(FCustomFieldDefs[i].TableName);
          WriteString(FCustomFieldDefs[i].FieldName);
          WriteString(FieldTypeToString(FCustomFieldDefs[i].FieldType));
          WriteInteger(FCustomFieldDefs[i].FieldSize);
          WriteInteger(FCustomFieldDefs[i].FieldPrecision);
          end;
       WriteListEnd;
       end;
end;

procedure TSqlitePassDatabaseDataTypeOptions.SetDefaultValues(DatabaseType: TSqlitePassDatabaseType; SetTranslationRules: Boolean=false);
begin
Case DatabaseType of
  dbtKexi: begin
           FDetectionMode := dmDbSpecific;
           FDateSeparator := '-';
           FDateStorage   := dsDbSpecific;
           FDecimalSeparator := '.';
           FDefaultFieldType := ftUnknown;
           FLongTimeFormat := '';
           FShortDateFormat := 'yyyy-mm-dd';
           FTimeSeparator := ':';
           FBooleanExtension := '';
           FLoadOptions := [soProperties, soTranslationRules, soCustomFieldDefs];
           FSaveOptions := [soProperties, soTranslationRules, soCustomFieldDefs];
           FDatabase.FOptions.FQuoteStyle := qsSingleQuote;
           end;

  dbtSqliteExpert:
           begin
           FDetectionMode := dmDbSpecific;
           FDateSeparator := '-';
           FDateStorage   := dsDbSpecific;
           FDecimalSeparator := '.';
           FDefaultFieldType := ftUnknown;
           FLongTimeFormat := '';
           FShortDateFormat := 'yyyy-mm-dd';
           FTimeSeparator := ':';
           FBooleanExtension := '';
           FLoadOptions := [soProperties, soTranslationRules, soCustomFieldDefs];
           FSaveOptions := [soProperties, soTranslationRules, soCustomFieldDefs];
           FDatabase.FOptions.FQuoteStyle := qsBracket;
           end;

  dbtSqlite4FPC:
           begin
           FDetectionMode := dmDbSpecific;
           FDateSeparator := '-';
           FDateStorage   := dsDbSpecific;
           FDefaultFieldType := ftString;
           FDecimalSeparator := '.';
           FLongTimeFormat := '';
           FShortDateFormat := 'yyyy-mm-dd';
           FTimeSeparator := ':';
           FBooleanExtension := '';
           FLoadOptions := [soProperties, soTranslationRules, soCustomFieldDefs];
           FSaveOptions := [soProperties, soTranslationRules, soCustomFieldDefs];
           FDatabase.FOptions.FQuoteStyle := qsSingleQuote;
           end;

  end;
if SetTranslationRules
   then TranslationRules.SetDefaultTranslationRules(DatabaseType);
end;

procedure TSqlitePassDatabaseDataTypeOptions.ApplyCustomFieldDefs;
var
  i: Integer;
  FieldDef: TSqlitePassTableFieldDef;
begin
for i:= 0 to Pred(FCustomFieldDefs.Count) do
    begin
    FieldDef := FDatabase.TableDefs.FindFieldDef(FCustomFieldDefs[i].TableName,FCustomFieldDefs[i].FieldName);
    if Assigned(FieldDef)
       then FieldDef.DataType := FCustomFieldDefs[i].FieldType;
    end;
end;
