{ This library is free software; you can redistribute it and/or modify it
  under the terms of the GNU Library General Public License as published by
  the Free Software Foundation; either version 2 of the License, or (at your
  option) any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Library General Public License
  for more details.

  You should have received a copy of the GNU Library General Public License
  along with this library; if not, write to the Free Software Foundation,
  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

  ---------------------------------------------------------------------------

    Author : Luc DAVID Email: luckylazarus@free.fr
    Copyright (C) 2006 - 2007

    Major changes are indicated in the \Documentation\Changes.pdf file
    Last update 04.09.2007

  --------------------------------------------------------------------------- }


{ TSqlitePassMemRecord }

{ --- Constructor / Destructor --- }
constructor TSqlitePassMemRecord.Create(FieldsCount: Integer);
begin
inherited Create;
SetLength(FieldsValues, FieldsCount);
end;

 { Using this destructor is not enough to set TSqlitePassMemRecord point to nil
   so we must always destroy a TSqlitePassMemRecord and set it to nil  }
destructor TSqlitePassMemRecord.Destroy(FieldDefs: TFieldDefs = nil);
begin
ClearAndFree(FieldDefs);
SetLength(FieldsValues, 0);
Inherited Destroy;
end;

procedure TSqlitePassMemRecord.Assign(const Source: TSqlitePassMemRecord; FieldDefs: TFieldDefs = nil);
begin
// TODO - Should RecordIndex be left to 0 ?
RecordIndex := Source.RecordIndex;
BookMarkFlag:= Source.BookMarkFlag;
AssignFieldsValue(Source, FieldDefs);
end;

procedure TSqlitePassMemRecord.AssignFieldsValue(const Source: TSqlitePassMemRecord; FieldDefs: TFieldDefs = nil);
Var
Idx: Integer;
BlobStream: TMemoryStream;
begin
For Idx := 0 to Pred(Length(FieldsValues)) do
begin
if Assigned(FieldDefs) and (Idx < FieldDefs.Count)
   then begin
        if (FieldDefs[Idx].DataType in SqlitePassBlobFieldTypes)
           and (Source.FieldsValues[Idx] <> '')
           then begin
                BlobStream := TMemoryStream.Create;
                BlobStream.LoadFromStream(TMemoryStream(StrToInt(Source.FieldsValues[Idx])));
                { We store the BlobStream adress in the Field }
                FieldsValues[Idx] := IntToStr(LongInt(BlobStream));
                end
            { Any other DataType }
            else FieldsValues[Idx] := Source.FieldsValues[Idx];
          end { Assigned(F) }
   { No fieldDefs }
   else FieldsValues[Idx] := Source.FieldsValues[Idx];
end; { for }
end;

procedure TSqlitePassMemRecord.ClearAndFree(FieldDefs: TFieldDefs = nil);
Var
Idx: Integer;
BlobStream: TMemoryStream;
begin
For Idx := 0 to Pred(Length(FieldsValues)) do
begin
if Assigned(FieldDefs) and (Idx < FieldDefs.Count) then
   begin
   if (FieldDefs[Idx].DataType in SqlitePassBlobFieldTypes)
      and (FieldsValues[Idx] <> '') then
          begin
          BlobStream := TMemoryStream(StrToInt(FieldsValues[Idx]));
          BlobStream.Free;
          end;
   end; { Assigned }
FieldsValues[Idx] := '';
end; { for }
end;


{ TSqlitePassRecordset}

constructor TSqlitePassRecordset.Create(Dataset: TSqlitePassDataset);
begin
  inherited Create;
  Self.Dataset := Dataset;
  { Minimum Capacity is set to 128.
  using a higher value when dealing with very large tables could improve speed }
  if Assigned(Dataset)
     then FFieldDefs := Dataset.FieldDefs
     else FFieldDefs := nil;
  Capacity := 128;
  CalcFieldsList := TList.Create;
end;

destructor TSqlitePassRecordset.Destroy;
begin
  Close;
  CalcFieldsList.Free;
  FFieldDefs := nil;
  Dataset := nil;
  inherited Destroy;
end;

{ --- Opening / Closing --- }

Function TSqlitePassRecordset.Open(DatabaseHandle: Pointer;
  Sql: String): Boolean;
begin
Result := True;
Close;
FDatabaseHandle := DatabaseHandle;
FSql := Sql;
if GetRecords
   then FActive := True
   else begin
        Close; { Safe call to free memory if needed }
        Result := False;
        end;
end;

procedure TSqlitePassRecordset.Close;
begin
if FActive then
   begin
   While Count > 0 do Delete(Pred(Count));

   if Assigned(FCurrentRecordOldValues)
     then FCurrentRecordOldValues.Destroy(FFieldDefs);

   ClearCalcFieldsList;
   FActive := False;
   end;
end;


{ --- Gets the records matching the SQL statement from sqlite engine --- }

function TSqlitePassRecordset.GetRecords: Boolean;
var
Accept: Boolean;
PDataArray: Pointer;
ColIdx, RowIdx, DataSize: Cardinal;
Rec: TSqlitePassMemRecord;
PRemainingSql: PChar;
i, ErrorCode, LowerLimit, UpperLimit, LimitIdx: Integer;
BlobStream: TMemoryStream;
Buffer: pointer;
begin
Result := True;

{ Retrieve data from sqlite }
if SqliteDbv3_prepare_v2(FDatabaseHandle, PChar(FSql), -1, PDataArray, PRemainingSql) <> SQLITE_OK
   then begin
        Result := False;
        Exit;
        end;

Try
DataFieldsCount := SqliteDbv3_column_count(PDataArray);

{ Set the Lower and Upper Limit to retrieve only the rows requested }
LimitIdx := 0;
LowerLimit := 1;
UpperLimit := Pred(MaxLongInt);

if Assigned(Dataset) then
   begin
   if Dataset.Filtered then
      begin
      LowerLimit := Max(1, Dataset.FilterRecordLowerLimit);
      if Dataset.FilterRecordUpperLimit <> 0
         then UpperLimit := Succ(Dataset.FilterRecordUpperLimit);
      end;

   InitCalcFieldsList(Dataset.Fields);
   SetFieldsNo(Dataset.Fields);
   end;

FieldsCount :=  DataFieldsCount + CalcFieldsList.Count;

{ We step through recordset until we reach LowerLimit or we get an error }
Repeat
  ErrorCode :=SqliteDbv3_Step(PDataArray);
  Inc(LimitIdx);
until (ErrorCode <> SQLITE_ROW) or (LimitIdx = LowerLimit);

RowIdx := 0;

{ Proceed with every records until we reach UpperLimit or we get an error }
While (ErrorCode = SQLITE_ROW) and (LimitIdx < UpperLimit) do
begin
Rec:=TSqlitePassMemRecord.Create(FieldsCount);

{ We get the fields values }
for ColIdx := 0 to Pred(DataFieldsCount) do
    begin
    if Assigned(Dataset) then
       begin
       if Dataset.FieldDefs[ColIdx].DataType in SqlitePassBlobFieldTypes
           then begin
                DataSize := SqliteDbv3_column_bytes(PDataArray,ColIdx);
                if Datasize > 0 then
                   begin
                   BlobStream := TMemoryStream.Create;
                   BlobStream.Write(SqliteDbv3_column_blob(PDataArray,ColIdx)^, DataSize);
                   { We store the BlobStream address in the Field }
                   Rec.FieldsValues[ColIdx] := IntToStr(LongInt(BlobStream));
                   end;
                end { ftBlob, ftMemo... }
            { Any other DataType is stored as a string }
            else Rec.FieldsValues[ColIdx] := String(SqliteDbv3_column_text(PDataArray,ColIdx));
       end { Assigned Dataset }
       else Rec.FieldsValues[ColIdx] := String(SqliteDbv3_column_text(PDataArray,ColIdx));
    end; { for }
Rec.RecordIndex := RowIdx;
Inc(RowIdx);
Inc(LimitIdx);
Add(Rec);

{ If needed, we send filter event now }
if Assigned(Dataset) then
   begin
   { dsFilter is set only when Dataset is Filtered and OnFilterRecord is Assigned }
   if Dataset.State = dsFilter then
      begin
      CurrentRecord := Rec;
      Dataset.OnFilterRecord(Dataset, Accept);
      CurrentRecord := nil;
      If Not Accept then
         begin
         Dec(RowIdx);
         Dec(LimitIdx);
         Delete(Pred(Count));
         end;
      end;
   end;

ErrorCode :=SqliteDbv3_Step(PDataArray);
end;

// TODO (FStatementTimeout=0) or ((FStatementTimeout>0) and (integer(GetTickCount)>=t));*)

{ Move to the first record }
First;

{ Create a utility record to be able to restore an edited record to
  his previous state if needed }
FCurrentRecordOldValues := TSqlitePassMemRecord.Create(FieldsCount);
finally
SqliteDbv3_finalize(PDataArray);
end;
end;

{ --- Navigating --- }

Procedure TSqlitePassRecordset.First;
begin
CurrentRecordIndex := -1;
end;

Procedure TSqlitePassRecordset.Prior;
begin
MoveBy(-1)
end;

Procedure TSqlitePassRecordset.Next;
begin
MoveBy(1)
end;

Procedure TSqlitePassRecordset.Last;
begin
CurrentRecordIndex := Count;
end;

Procedure TSqlitePassRecordset.MoveBy(Offset: Integer);
begin
Try
MoveState := grOk;
Case Offset of
   { Move to Previous record }
   -1 : if CurrentRecordIndex < 1
           then MoveState := grBOF
           else CurrentRecordIndex := CurrentRecordIndex -1;
   { Current record }
    0 : if (CurrentRecordIndex < 0) or (CurrentRecordIndex > Pred(Count))
           then MoveState := grError
                { This will update CurrentRecord }
           else CurrentRecordIndex := CurrentRecordIndex;
   { Move to the Next record }
    1 : if (CurrentRecordIndex >= Pred(Count))
           then MoveState := grEOF
           else CurrentRecordIndex := CurrentRecordIndex +1;
    end;
Except
MoveState := grError
end;
end;


{ --- Setting current record index or current record --- }

procedure TSqlitePassRecordset.SetFCurrentRecordIndex(
  const Value: Integer);
begin
FCurrentRecord := nil;
if Value < 0
   then FCurrentRecordIndex := -1 { BOF }
   else if Value >= Count
           then FCurrentRecordIndex := Count { EOF }
           else begin
                FCurrentRecordIndex := Value;
                FCurrentRecord := GetItems(Value); { We get the current record index}
                end;
end;

procedure TSqlitePassRecordset.SetFCurrentRecord(const Value: TSqlitePassMemRecord);
begin
FCurrentRecord := Value;
if FCurrentRecord <> nil
   then FCurrentRecordIndex := FCurrentRecord.RecordIndex
   else FCurrentRecordIndex := -1;
end;


{ --- Managing Records --- }

procedure TSqlitePassRecordset.AddRecord;
begin
if NewRecord.RecordIndex = Count
   then Add(NewRecord)
   else Insert(NewRecord.RecordIndex, NewRecord);
{ NewRecord adress is now kept by the recordlist and must be reset to nil }
NewRecord := nil;
end;

Function TSqlitePassRecordset.Add(ARecord: TSqlitePassMemRecord): Integer;
begin
  Result:=Inherited Add(ARecord);
end;

Procedure TSqlitePassRecordset.Insert(Index: Integer; ARecord: TSqlitePassMemRecord);
begin
  Inherited Insert(Index,ARecord);
  UpdateRecordIndex(Index);
end;

procedure TSqlitePassRecordset.Cancel;
begin
  { Move to the last selected record }
  CurrentRecordIndex := PreviousRecordIndex;
  NewRecord.Destroy(FFieldDefs);
  { We need to set NewRecord to nil otherwise it will be pointing to undefined memory }
  NewRecord := nil;
end;

Function TSqlitePassRecordset.GetItems(Index: Integer): TSqlitePassMemRecord;
begin
  Result:=TSqlitePassMemRecord(Inherited Items[Index]);
end;

Procedure TSqlitePassRecordset.SetItems(Index: Integer; ARecord: TSqlitePassMemRecord);
begin
  Put(Index,ARecord);
end;

procedure TSqlitePassRecordset.Delete(Index: Integer);
begin
   { Clear record content and free blobs }
  Items[Index].Destroy(FFieldDefs);
  Inherited Delete(Index);
  CurrentRecordIndex := Index;
  UpdateRecordIndex(Index);
end;

{ --- Update the TSqlitePassMemRecord.RecordIndex Value when
      inserting or deleting a new record   --- }
procedure TSqlitePassRecordset.UpdateRecordIndex(StartIndex: integer = 0);
var
i: Integer;
begin
for i := StartIndex to Pred(Count)
    do Items[i].RecordIndex := i;
end;

{ --- Managing Records --- }

function TSqlitePassRecordset.CreateNewRecord(Fields: TFields): TSqlitePassMemRecord;
Var
i, FieldNo: Integer;
begin
NewRecord := TSqlitePassMemRecord.Create(FieldsCount);

{ Set default values if needed }
For i := 0 to Pred(FieldsCount)
    do begin
       If i < Fields.Count then
          begin
          if Fields[i].FieldKind = fkCalculated
             then FieldNo := 0//GetCalcFieldNo(Fields[i].FieldName)
             else FieldNo := Fields[i].FieldNo-1;
          NewRecord.FieldsValues[FieldNo] := Fields[i].DefaultExpression;
          end;
       end;

NewRecord.RecordIndex := Count;
Result := NewRecord;
end;

{ --- Current record old values backup / restore when editing --- }

procedure TSqlitePassRecordset.SaveCurrentRecordOldValues;
begin
FCurrentRecordOldValues.AssignFieldsValue(CurrentRecord, Dataset.FieldDefs);
end;
{cv 1,2
ov 3,4
cv 1,2 free
cv 5,6 create
ov 3,4 free}
procedure TSqlitePassRecordset.RestoreCurrentRecordOldValues;
begin
CurrentRecord.ClearAndFree(Dataset.FieldDefs);
CurrentRecord.AssignFieldsValue(FCurrentRecordOldValues);
FCurrentRecordOldValues.ClearAndFree;
end;

{ Populate the CalcFielList }
procedure TSqlitePassRecordset.InitCalcFieldsList(Fields: TFields);
var
i, Idx: Integer;
CalcField: PCalcField;
begin
Idx := DataFieldsCount;
for i := 0 to Pred(Fields.Count) do
 if Fields[i].FieldKind = fkCalculated
       then begin
            New(CalcField);
            CalcField^.FieldName := Fields[i].FieldName;
            CalcField^.FieldNo := Idx;
            CalcFieldsList.Add(CalcField);
            Inc(Idx);
            end;
end;

{ Populate the CalcFielList }
procedure TSqlitePassRecordset.ClearCalcFieldsList;
var
CalcField: PCalcField;
begin
While CalcFieldsList.Count > 0 do
      begin
      CalcField := CalcFieldsList.Items[Pred(CalcFieldsList.Count)];
      CalcFieldsList.Delete(Pred(CalcFieldsList.Count));
      Dispose(CalcField);
      end;
end;

function TSqlitePassRecordset.GetCalcFieldNo(
  FieldName: String): Integer;
var
i: Integer;
begin
Result := -1;
For i := 0 to Pred(CalcFieldsList.Count) do
    if PCalcField(CalcFieldsList.Items[i])^.FieldName = FieldName
       then begin
            Result := PCalcField(CalcFieldsList.Items[i])^.FieldNo;
            Exit;
            end;
end;


procedure TSqlitePassRecordset.ClearCalcFields(ARecord: TSqlitePassMemRecord);
var
i: Integer;
begin
For i := DataFieldsCount to Pred(FieldsCount)
    do ARecord.FieldsValues[i] := '';
end;

{ Returns the internal value position for a given field.index }
function TSqlitePassRecordset.GetFieldNo(Index: Integer): Integer;
begin
Result := FFieldsNo[Index];
end;


procedure TSqlitePassRecordset.SetFieldsNo(Fields: TFields);
var
i, CalcFieldsOffset: Integer;
begin
if (csLoading in Dataset.ComponentState) then exit;
SetLength(FFieldsNo, Fields.Count);

if CalcFieldsList.Count > 0
   then CalcFieldsOffset := CalcFieldsList.Count
   else CalcFieldsOffset := 1;

for i := 0 to Pred(Fields.Count) do
    begin
    Case Fields[i].FieldKind of
         fkData: FFieldsNo[i] := Fields[i].FieldNo - CalcFieldsOffset;
         fkCalculated: FFieldsNo[i] := GetCalcFieldNo(Fields[i].FieldName);
         end;
    end;
end;



